# instrument_base.py, base class to implement instrument objects
# Reinier Heeres <reinier@heeres.eu>, 2008
# HR@KIT 2017 (converted to python3)
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from __future__ import print_function

import copy
import inspect
import logging
import time

import numpy as np

import qkit


class Instrument(object):
    """
    Base class for instruments.

    Usage:
    Instrument.get(<variable name or list>)
    Instrument.set(<variable name or list>)

    Implement an instrument:
    In __init__ call self.add_variable(<name>, <option dict>)
    Implement _do_get_<variable> and _do_set_<variable> functions
    """



    FLAG_GET = 0x01             # Parameter is gettable
    FLAG_SET = 0x02             # {arameter is settable
    FLAG_GETSET = 0x03          # Shortcut for GET and SET
    FLAG_GET_AFTER_SET = 0x04   # perform a 'get' after a 'set'
    FLAG_SOFTGET = 0x08         # 'get' operation is simulated in software,
                                # e.g. an internally stored value is returned.
                                # Only use for parameters that cannot be read
                                # back from a device.

    RESERVED_NAMES = ('name', 'type')

    def __init__(self, name, **kwargs):

        self._name = name
        self._initialized = False

        self._options = kwargs
        if 'tags' not in self._options:
            self._options['tags'] = []

        self._parameters = {}
        self._parameter_groups = {}
        self._functions = {}
        self._added_methods = []
        self._probe_ids = []
        self._offsets = {}

    def __str__(self):
        return "Instrument '%s'" % (self.get_name())


    def get_name(self):
        '''
        Returns the name of the instrument as a string

        Input: None
        Output: name of instrument (string)
        '''

        return self._name
    
    def _json(self):
        '''
        This function is called by the QkitJSONEncoder and should return a 
        dictionary that will be saved to the JSON file.
        '''
        return {'dtype' : 'qkitInstrument', 'content': str(self.get_name())}
        
    def get_type(self):
        """Return type of instrument as a string."""
        modname = str(self.__module__)
        return modname

    def get_options(self):
        '''Return instrument options.'''
        return self._options

    def get_tags(self):
        '''
        Returns array of tags

        Input: None
        Output: array of strings
        '''
        return self._options['tags']

    def add_tag(self, tag):
        '''
        Add tag to the tag list

        Input: tag (string)
        Output: None
        '''
        self._options['tags'].append(tag)

    def has_tag(self, tags):
        '''
        Return whether instrument has any tag in 'tags'
        '''

        if type(tags) is list:
            for tag in tags:
                if tag in self._options['tags']:
                    return True
            return False

        else:
            if tags in self._options['tags']:
                return True

        return False

    def initialize(self):
        '''
        Currently unsupported; might be used in the future to perform
        extra initialization in sub-classed Instruments.

        Input: None
        Output: None
        '''
        self._initialized = True

    def remove(self):
        '''
        Notify the instrument collection that this instrument should be
        removed. Override this in a sub-classed Instrument to perform
        cleanup.

        Input: None
        Output: None
        '''

        self._remove_parameters()

    def is_initialized(self):
        '''
        Return whether Instrument is initialized.

        Input: None
        Output: Boolean
        '''
        return self._initialized

    def add_parameter(self, name, **kwargs):
        '''
        Create an instrument 'parameter' that is known by the whole
        environment (gui etc).

        This function creates the 'get_<name>' and 'set_<name>' wrapper
        functions that will perform checks on parameters and finally call
        '_do_get_<name>' and '_do_set_<name>'. The latter functions should
        be implemented in the instrument driver.

        Input:
            name (string): the name of the parameter (string)
            optional keywords:
                type: types.FloatType, types.StringType, etc.
                flags: bitwise or of Instrument.FLAG_ constants.
                    If not set, FLAG_GETSET is default
                channels: tuple. Automagically create channels, e.g.
                    (1, 4) will make channels 1, 2, 3, 4.
                minval, maxval: values for bound checking
                units (string): units for this parameter
                maxstep (float): maximum step size when changing parameter
                stepdelay (float): delay when setting steps (in milliseconds)
                tags (array): tags for this parameter
                doc (string): documentation string to add to get/set functions
                listen_to (list of (ins, param) tuples): list of parameters
                    to watch. If any of them changes, execute a get for this
                    parameter. Useful for a parameter that depends on one
                    (or more) other parameters.

        Output: None
        '''

        if name in self._parameters:
            logging.error('Parameter %s already exists.', name)
            return
        elif name in self.RESERVED_NAMES:
            logging.error("'%s' is a reserved name, not adding parameter",
                    name)
            return

        options = kwargs
        if 'flags' not in options:
            options['flags'] = Instrument.FLAG_GETSET
        if 'type' not in options:
            options['type'] = type(None)
        if 'tags' not in options:
            options['tags'] = []
        if 'offset' not in options:
            options['offset'] = False

        # If defining channels call add_parameter for each channel
        if 'channels' in options:
            if len(options['channels']) == 2 and type(options['channels'][0]) is int:
                minch, maxch = options['channels']
                channels = range(minch, maxch + 1)
            else:
                channels = options['channels']

            for i in channels:
                chopt = copy.copy(options)
                del chopt['channels']
                chopt['channel'] = i
                chopt['base_name'] = name

                if 'channel_prefix' in options:
                    var_name = options['channel_prefix'] % i + name
                else:
                    var_name = '%s%s' % (name, i)

                self.add_parameter(var_name, **chopt)

            return

        self._parameters[name] = options
        self._offsets[name] = None

        if 'channel' in options:
            ch = options['channel']
        else:
            ch = None

        base_name = kwargs.get('base_name', name)

        if options['offset']:
            if options["type"] in (bool,bytes,str):
                raise ValueError("%s.%s: Offset not available for parameter with type %s" % (self._name, name, options["type"].__name__))
            else:
                func = lambda value: self._offsets.update({name:value})
                func.__doc__ = """Set an offset for parameter %s on device %s.
                The offset is added on every get and substracted on every set command.
                Use e.g. set_power_offset(-20) if you add 20dB attenuation at the device."""%(self._name,name)
                setattr(self, 'set_%s_offset' % name, func)
                self._added_methods.append('set_%s_offset' % name)
    
                setattr(self, 'get_%s_offset' % name, lambda : self._offsets[name])
                self._added_methods.append('get_%s_offset' % name)

        if options['flags'] & Instrument.FLAG_GET:
            if ch is not None:
                func = lambda query=True, **lopts: \
                    self.get(name, query=query, channel=ch, **lopts)
            else:
                func = lambda query=True, **lopts: \
                    self.get(name, query=query, **lopts)

            func.__doc__ = 'Get variable %s' % name
            if 'doc' in options:
                func.__doc__ += '\n%s' % options['doc']

            setattr(self, 'get_%s' % name,  func)
            self._added_methods.append('get_%s' % name)

            # Set function to do_get_%s or _do_get_%s, whichever is available
            # (if no function specified)
            if 'get_func' not in options:
                options['get_func'] = getattr(self, 'do_get_%s' % base_name, \
                    getattr(self, '_do_get_%s' % base_name, None))
            if options['get_func'] is not None:
                if options['get_func'].__doc__ is not None:
                    func.__doc__ += '\n%s' % options['get_func'].__doc__
            else:
                options['get_func'] = lambda *a, **kw: \
                    self._get_not_implemented(base_name)
                self._get_not_implemented(base_name)

        if options['flags'] & Instrument.FLAG_SOFTGET:
            if ch is not None:
                func = lambda query=True, **lopts: \
                    self.get(name, query=False, channel=ch, **lopts)
            else:
                func = lambda query=True, **lopts: \
                    self.get(name, query=False, **lopts)

            func.__doc__ = 'Get variable %s (internal stored value)' % name
            setattr(self, 'get_%s' % name,  func)
            self._added_methods.append('get_%s' % name)

        if options['flags'] & Instrument.FLAG_SET:
            if ch is not None:
                func = lambda val, **lopts: self.set(name, val, channel=ch, **lopts)
            else:
                func = lambda val, **lopts: self.set(name, val, **lopts)

            func.__doc__ = 'Set variable %s' % name
            if 'doc' in options:
                func.__doc__ += '\n%s' % options['doc']
            setattr(self, 'set_%s' % name, func)
            self._added_methods.append('set_%s' % name)

            # Set function to do_set_%s or _do_set_%s, whichever is available
            # (if no function specified)
            if 'set_func' not in options:
                options['set_func'] = getattr(self, 'do_set_%s' % base_name, \
                    getattr(self, '_do_set_%s' % base_name, None))
            if options['set_func'] is not None:
                if options['set_func'].__doc__ is not None:
                    func.__doc__ += '\n%s' % options['set_func'].__doc__
            else:
                options['set_func'] = lambda *a, **kw: \
                    self._set_not_implemented(base_name)
                self._set_not_implemented(base_name)

#        setattr(self, name,
#            property(lambda: self.get(name), lambda x: self.set(name, x)))

        if 'listen_to' in options:
            insset = set([])
            inshids = []
            for (ins, param) in options['listen_to']:
                inshids.append(ins.connect('changed', \
                        self._listen_parameter_changed_cb,
                        param, options['get_func']))
            options['listed_hids'] = inshids

        if 'group' in options:
            g = options['group']
            if g not in self._parameter_groups:
                self._parameter_groups[g] = [name]
            else:
                self._parameter_groups[g].append(name)

    def _remove_parameters(self):
        '''
        Remove remaining references to bound methods so that the Instrument
        object can be garbage collected.
        '''

        for name, opts in self._parameters.items():
            for fname in ('get_%s' % name, 'set_%s' % name):
                if hasattr(self, fname):
                    delattr(self, fname)
        self._parameters = {}

    def remove_parameter(self, name):
        if name not in self._parameters:
            return

        for func in ('get_%s' % name, 'set_%s' % name):
            if hasattr(self, func):
                delattr(self, func)

        del self._parameters[name]

    def has_parameter(self, name):
        '''
        Return whether instrument has a parameter called 'name'
        '''
        return (name in self._parameters)

    def get_parameter_options(self, name):
        '''
        Return list of options for paramter.

        Input: name (string)
        Output: dictionary of options
        '''
        if name in self._parameters:
            return self._parameters[name]
        else:
            return None

    def get_shared_parameter_options(self, name):
        '''
        Return list of options for paramter.

        Input: name (string)
        Output: dictionary of options
        '''
        if name in self._parameters:
            options = dict(self._parameters[name])
            for i in ('get_func', 'set_func'):
                if i in options:
                    del options[i]
            if 'type' in options and options['type'] is type(None):
                options['type'] = None
            return options
        else:
            return None

    def set_parameter_options(self, name, **kwargs):
        '''
        Change parameter options.

        Input:  name of parameter (string)
        Ouput:  None
        '''
        if name not in self._parameters:
            print('Parameter %s not defined' % name)
            return None

        for key, val in kwargs.items():
            self._parameters[name][key] = val

    def get_parameter_tags(self, name):
        '''
        Return tags for parameter 'name'.

        Input:  name of parameter (string)
        Ouput:  array of tags
        '''

        if name not in self._parameters:
            return []

        return self._parameters[name]['tags']

    def add_parameter_tag(self, name, tag):
        '''
        Add tag to list of tags for parameter 'name'.

        Input:  (1) name of parameter (string)
                (2) tag (string)
        Ouput:  None
        '''

        if name not in self._parameters:
            return

        self._parameters[name]['tags'].append(tag)

    def set_parameter_bounds(self, name, minval, maxval):
        '''
        Change the bounds for a parameter.

        Input:  (1) name of parameter (string)
                (2) minimum value
                (3) maximum value
        Output: None
        '''
        self.set_parameter_options(name, minval=minval, maxval=maxval)

    def set_channel_bounds(self, name, channel, minval, maxval):
        '''
        Change the bounds for a channel.

        Input:  (1) name of parameter (string)
                (2) channel number (int)
                (3) minimum value
                (4) maximum value
        Output: None
        '''

        opts = self.get_parameter_options(name)
        if 'channel_prefix' in opts:
            var_name = opts['channel_prefix'] % channel + name
        else:
            var_name = '%s%d' % (name, channel)

        self.set_parameter_options(var_name, minval=minval, maxval=maxval)

    def set_parameter_rate(self, name, stepsize, stepdelay):
        '''
        Change the rate properties for a channel.

        Input:
            stepsize (float): the maximum step size
            stepdelay (float): the delay after each step
        Output:
            None
        '''
        self.set_parameter_options(name, maxstep=stepsize, stepdelay=stepdelay)

    def get_parameter_names(self):
        '''
        Returns a list of parameter names.

        Input: None
        Output: all the paramter names (list of strings)
        '''
        return self._parameters.keys()

    def get_parameters(self):
        '''
        Return the parameter dictionary.

        Input: None
        Ouput: Dictionary, keys are parameter names, values are the options.
        '''
        return self._parameters

    def get_shared_parameters(self):
        '''
        Return the parameter dictionary, with non-shareable items stripped.

        Input: None
        Ouput: Dictionary, keys are parameter names, values are the options.
        '''

        params = {}
        for key in self._parameters:
            params[key] = self.get_shared_parameter_options(key)
        return params

    def get_parameter_groups(self):
        '''
        Return a dictionary with parameter group name -> group members.
        '''
        return self._parameter_groups
    
    def _offset(self,name,value,sign):
        if self._offsets[name] is not None:
            return value+sign*self._offsets[name]
        return value

    def _get_value(self, name, query=True, **kwargs):
        '''
        Private wrapper function to get a value.

        Input:  (1) name of parameter (string)
                (2) query the instrument or return stored value (Boolean)
                (3) optional list of extra options
        Output: value of parameter (whatever type the instrument driver returns)
        '''

        try:
            p = self._parameters[name]
        except:
            print('Could not retrieve options for parameter %s' % name)
            return None

        if 'channel' in p and 'channel' not in kwargs:
            kwargs['channel'] = p['channel']

        flags = p['flags']
        if not query or flags & 8: #self.FLAG_SOFTGET:
            if 'value' in p:
                if p['type'] == np.ndarray:
                    return self._offset(name,np.array(p['value']),+1)
                else:
                    return self._offset(name,p['value'],+1)
            elif not query:
                if not flags & 1: #Instrument.FLAG_GET, not gettable
                    return None 
                # if this value is not stored but gettable, don't throw an error but just get it.
                logging.debug("%s.%s value not cached. I try getting it with query=True."%(self._name,name))
                return self._get_value(name,query=True,**kwargs)
            else:
                logging.error('Trying to access cached value %s, but none available'%name)
                return False

        # Check this here; getting of cached values should work
        if not flags & 1: #Instrument.FLAG_GET:
            if query:
                logging.error('Instrument %s does not support getting of %s' %(self._name, name))
            return None

        func = p['get_func']
        value = func(**kwargs)
        if 'type' in p and value is not None:
            try:
                if p['type'] == bytes or p['type'] == type(None):
                    pass
                elif p['type'] == np.ndarray:
                    value = np.array(value)
                else:
                    value = p['type'](value)
            except:
                logging.warning('Unable to cast value "%s" to %s', value, p['type'])

        p['value'] = value
        return self._offset(name,value,+1)

    def get(self, name, query=True, fast=False, **kwargs):
        '''
        Get one or more Instrument parameter values.

        Input:
            name (string or list/tuple of strings): name of parameter(s)
            query (bool): whether to query the instrument or return the
                last stored value
            fast (bool): legacy from qtlab
            kwargs: Optional keyword args that will be passed on.

        Output: Single value, or dictionary of parameter -> values
                Type is whatever the instrument driver returns.
        '''
        if type(name) in (list, tuple):
            result = {}
            for key in name:
                val = self._get_value(key, query, **kwargs)
                if val is not None:
                    result[key] = val
        else:
            result = self._get_value(name, query, **kwargs)
        qkit.flow.sleep()
        return result

    def get_threaded(self, *args, **kwargs):
        logging.error('Using threading functions is not supported. Redirecting to normal get')
        return self.get(*args, **kwargs)

    _CONVERT_MAP = {
            int: int,
            float: float,
            bytes: str,
            str: str,
            bool: bool,
            tuple: tuple,
            list: list,
            np.ndarray: lambda x: x.tolist(),
    }

    def _convert_value(self, value, ttype):
        if type(value) is bool and \
                ttype is not bool:
            raise ValueError('Setting a boolean, but that is not the expected type')

        if ttype not in self._CONVERT_MAP:
            raise ValueError('Unsupported type %s', ttype)

        try:
            func = self._CONVERT_MAP[ttype]
            value = func(value)
        except:
            raise ValueError('Conversion of %r to type %s failed',value, ttype)

        return value

    def _set_value(self, name, value, **kwargs):
        '''
        Private wrapper function to set a value.

        Input:  (1) name of parameter (string)
                (2) value of parameter (whatever type the parameter supports).
                    Type casting is performed if necessary.
                (3) Optional keyword args that will be passed on.
        Output: Value returned by the _do_set_<name> function,
                or result of get in FLAG_GET_AFTER_SET specified.
        '''
        if name in self._parameters:
            p = self._parameters[name]
        else:
            raise ValueError("Parameter %s not known"%name)

        if not p['flags'] & Instrument.FLAG_SET:
            print('Instrument does not support setting of %s' % name)
            return None

        if 'channel' in p and 'channel' not in kwargs:
            kwargs['channel'] = p['channel']

        value = self._offset(name, value, -1)

        if 'type' in p:
            value = self._convert_value(value, p['type'])

        if 'minval' in p and value < p['minval']:
            if self._offsets[name] is not None and self._offsets[name] != 0:
                raise qkit.instruments.InstrumentBoundsError('Cannot set %s.%s to %g: With offset %g, %s at %s would be %g, which is too small (Minimum: '
                                                             '%g)' % (self._name, name, value+self._offsets[name],self._offsets[name],name,self._name,value, p['maxval']))
            raise qkit.instruments.InstrumentBoundsError('Cannot set %s.%s to %s: value too small (Minimum: %g)' % (self._name, name,value, p['minval']))

        if 'maxval' in p and value > p['maxval']:
            if self._offsets[name] is not None and self._offsets[name] != 0:
                raise qkit.instruments.InstrumentBoundsError('Cannot set %s.%s to %g: With offset %g, %s at %s would be %g, which is too large (Maximum: '
                                                             '%g)' % (self._name, name, value+self._offsets[name],self._offsets[name],name,self._name,value, p['maxval']))
            raise qkit.instruments.InstrumentBoundsError('Cannot set %s.%s to %s: value too large (Maximum: %g)' % (self._name, name, value, p['maxval']))

        func = p['set_func']
        if 'maxstep' in p and p['maxstep'] is not None:
            curval = p['value']
            if curval is None:
                logging.warning('Current value not available, ignoring maxstep')
                curval = value + 0.01 * p['maxstep']

            sign = np.sign(value - curval)
            for v in np.arange(curval, value, sign * np.abs(p['maxstep'])):
                func(v, **kwargs)  # execute the set function
                time.sleep(p.get('stepdelay', 50) / 1000.)

        func(value, **kwargs)  # execute the set function to get the final value

        if p['flags'] & self.FLAG_GET_AFTER_SET:
            newvalue = self._offset(name,self._get_value(name, **kwargs),-1)
            if newvalue != value:
                logging.warning("%s.%s: actual value (%s) differs from set value (%s)"%(self._name,name,newvalue,value))
            value = newvalue

        p['value'] = value
        return value

    def set(self, name, value=None, fast=False, **kwargs):
        '''
        Set one or more Instrument parameter values.

        Checks whether the Instrument is locked and checks value bounds,
        if specified by minval / maxval.

        Input:
            name (string or dict): which parameter to set, or dictionary of
                parameter -> value
            value (any): the value to set
            fast (bool): if True perform as fast as possible, e.g. don't
                emit a signal to update the GUI.
            kwargs: Optional keyword args that will be passed on.

        Output: True or False whether the operation succeeded.
                For multiple sets return False if any of the parameters failed.
        '''

        result = True
        changed = {}
        if type(name) == dict:
            for key, val in name.items():
                val = self._set_value(key, val, **kwargs)
                if val is not None:
                    changed[key] = val
                else:
                    result = False

        else:
            val = self._set_value(name, value, **kwargs)
            if val is not None:
                changed[name] = val
            else:
                result = False

        qkit.flow.sleep()
        return result

    def get_argspec_dict(self, a):
        return dict(args=a[0], varargs=a[1], keywords=a[2], defaults=a[3])

    def add_function(self, name, **options):
        '''
        Inform the Instrument wrapper class to expose a function.

        Input:  (1) name of function (string)
                (2) dictionary of extra options
        Output: None
        '''

        if not hasattr(self, name):
            logging.warning('Instrument does not implement function %s', name)

        f = getattr(self, name)
        if hasattr(f, '__doc__'):
            options['doc'] = getattr(f, '__doc__')

        options['argspec'] = self.get_argspec_dict(inspect.getfullargspec(f))

        self._functions[name] = options

    def get_function_options(self, name):
        '''
        Return options for an Instrument function.

        Input:  name of function (string)
        Output: dictionary of options for function 'name'
        '''
        if name in self._functions:
            return self._functions[name]
        else:
            return None

    def get_function_parameters(self, name):
        '''
        Return info about parameters for function.
        '''
        if name in self._functions:
            if 'parameters' in self._functions[name]:
                return self._functions[name]['parameters']
        return None

    def get_function_names(self):
        '''
        Return the list of exposed Instrument function names.

        Input: None
        Output: list of function names (list of strings)
        '''
        return self._functions.keys()

    def get_functions(self):
        '''
        Return the exposed functions dictionary.

        Input: None
        Ouput: Dictionary, keys are function  names, values are the options.
        '''
        return self._functions

    def call(self, funcname, **kwargs):
        '''
        Call the exposed function 'funcname'.

        Input:  (1) function name (string)
                (2) Optional keyword args that will be passed on.
        Output: None
        '''
        f = getattr(self, funcname)
        f(**kwargs)

    def reload(self):
        '''
        Reloads the instrument. Make sure to only use the new instance
        of the instrument, which is returned here!

        Input:
            None
        Output:
            New instance of instrument
        '''
        logging.warning("No warranty on reloads. Make sure to re-assign the instrument as \n %s = %s.reload()"%(self._name,self._name))
        return qkit.instruments.reload(self)

    def _get_not_implemented(self, name):
        logging.warning('Get not implemented for %s.%s' % \
            (Instrument.get_type(self), name))

    def _set_not_implemented(self, name):
        logging.warning('Set not implemented for %s.%s' % \
            (Instrument.get_type(self), name))

    def _listen_parameter_changed_cb(self, sender, changed,listen_param, update_func):
        # ToDo this has to be correctly implemented!
        # Cross-related parameters are probably a good thing to have
        # i.e. one parameter is "getted" if another one is set
        if listen_param not in changed:
            return
        update_func()

class InvalidInstrument(Instrument):
    '''
    Placeholder class for instruments that fail to load, mainly to support
    reloading.
    '''

    def __init__(self, name, instype, **kwargs):
        self._instype = instype
        self._kwargs = kwargs
        Instrument.__init__(self, name)

    def get_type(self):
        return self._instype

    def get_create_kwargs(self):
        return self._kwargs

class GPIBInstrument(Instrument):
    def __init__(self, *args, **kwargs):
        kwargs['lockclass'] = 'GPIB'
        Instrument.__init__(self, *args, **kwargs)



#!/usr/bin/env python
"""
GPIB_ETHERNET python class for prologix ethernet-to-gpib bridge, 
written by Hannes Rotzinger, hannes.rotzinger@kit.edu for qtlab, qtlab.sf.net

interfaces GPIB commands over ethernet

released under the GPL, whatever version

Changelog:
0.1 March 2010, initial version, very alpha, most of the functionality is far from bullet proof.

0.2 June 2012 added master/slave capability and multi-device support -- by Jochen Zimmer

"""

import socket
import time
import re

class instrument(object):
    """ for prologix gpib_ethernet bridge """

    N_prologix_adapters = 0  # number of adapters
    l_IPs = []         # IPs of the adapters controlled by this driver
    l_p_master = []    # master object for adapter i

    def __init__(self,gpib,**kwargs):

        self.sock = None
        
        # for compatibility with NI visa
        self.timeout = kwargs.get("timeout",5)
        self.chunk_size = kwargs.get("chunk_size",20*1024)
        self.values_format = kwargs.get("values_format",'ascii') # fixme: single, double
        self.term_char = kwargs.get("term_char",'\n')
        self.send_end = kwargs.get("send_end",True)
        self.delay = kwargs.get("delay",0)
        self.lock = kwargs.get("lock",False)
        
        # parse gpib address (throws an Error() if fails)
        self.gpib_addr=self._get_gpib_adr_from_string(gpib)
        
        # IP address and port of  PROLOGIX GPIB-ETHERNET
        self.ip = kwargs.get("ip","192.168.0.100")
        self.ethernet_port = kwargs.get("port", 1234)

        self.p_master = None

        for i in range(instrument.N_prologix_adapters):
            if self.ip == instrument.l_IPs[i]:
                # print "Found a slave: %s, %i" % (self.ip, self.gpib_addr)
                self.is_master = False
                self.p_master = instrument.l_p_master[i]
        if self.p_master == None:
            # print "Found a master: %s, %i" % (self.ip, self.gpib_addr)
            instrument.N_prologix_adapters += 1
            instrument.l_IPs.append(self.ip)
            self.is_master = True
            instrument.l_p_master.append(self)
            self.p_master = self

#       self.gpib_addr

        if self.is_master:
            # open connection
            self._open_connection()

            if self.send_end:
                    self.term_char = '\n'
                    self._set_EOI_assert()

            # disable the automatic saving of parameters in
            # the ethernet-gpib device,

            #self._set_saveconfig(False) ## deactivated because it caused an unknown error in the new devices
            
            # set set the ethernet gpib device to be the controller of#
            # the gpib chain
            self._set_controller_mode()
            # set the GPIB address of the device
            self._set_gpib_address()
            
            self._set_EOI_assert()
            self._set_read_timeout()
            self._set_read_after_write(False)
            #self._dump_internal_vars()

    # wrapper functions for py visa, generalized for master/slave
    def write(self,cmd):
        self.p_master._send("++addr " + str(self.gpib_addr) + "\n")
        return self.p_master._send(cmd)

    def read(self):
        self.p_master._send("++addr " + str(self.gpib_addr) + "\n")
        return self.p_master._recv()

    def read_values(self,format):
        self.p_master._send("++addr " + str(self.gpib_addr) + "\n")
        return self.p_master._recv()
        
    def ask(self,cmd):
        self.p_master._send("++addr " + str(self.gpib_addr) + "\n")
        return self.p_master._send_recv(cmd)

    def ask_for_values(self,cmd,format=None):
        self.p_master._send("++addr " + str(self.gpib_addr) + "\n")
        return self.p_master._send_recv(cmd)
        
    def clear(self):
        self.p_master._send("++addr " + str(self.gpib_addr) + "\n")
        return self.p_master._set_reset()

    def trigger(self):
        self.p_master._send("++addr " + str(self.gpib_addr) + "\n")
        return self.p_master._set_trigger()


    # this is in the Gpib class of pyvisa, has to move there later
    def send_ifc(self):
        return self._set_ifc()

    # utility functions
    def _get_gpib_adr_from_string(self,gpib_str):
        # very,very simple GPIB address extraction.
        p = re.compile('(gpib::|GPIB::)(\d+)')
        m = p.match(gpib_str)
        if m:
            return int(m.group(2))
        else:
            raise self.Error("Only GPIB:: is supported!")
    # 
    # internal commands to access the prologix gpib device
    #
    
    def _send(self,cmd):
        cmd=cmd.rstrip()
        cmd+=self.term_char
        #self._set_read_after_write(False)
        self.sock.send(cmd.encode())
        time.sleep(self.delay)
        
    def _send_recv(self,cmd,**kwargs):
        bufflen=kwargs.get("bufflen",self.chunk_size)
        cmd=cmd.rstrip()
        cmd+=self.term_char
        #print cmd
        
        self._send(cmd)
        self._set_read()

        response=''
        while True:
            resp=self.sock.recv(bufflen).decode()
            if resp == '':
                raise RuntimeError("socket connection broken")
            response+=resp
            if response[-1] == '\n': # kind of delimeter used to check if message is complete, problem when response too short
                break
        return response

    def _recv(self, **kwargs):
        bufflen = kwargs.get("bufflen", self.chunk_size)
        response=''
        while True:
            resp=self.sock.recv(bufflen).decode()
            if resp == '':
                raise RuntimeError("socket connection broken")
            response+=resp
            if response[-1] == '\n': # kind of delimeter used to check if message is complete, problem when response too short
                break
        return response


    def _open_connection(self,**kwargs):
        # these calls should not be used any more
        self.ip = kwargs.get("ip",self.ip)
        #self.gpib_addr = kwargs.get("gpib_addr",self.gpib_addr)
        # Open TCP connect to port 1234 of GPIB-ETHERNET
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM) #socket.IPPROTO_TCP)
        self.sock.settimeout(3)
        self.sock.connect((self.ip, self.ethernet_port))

    def _close_connection(self):
        self.sock.close()

    def _set_read(self):
        self._send("++read eoi")

    def _set_saveconfig(self,On=False):
        # should not be used very frequently
        if On:
            self._send("++savecfg 1")
        else:
            self._send("++savecfg 0")

    def _set_gpib_address(self,**kwargs):
        # GET GPIB address
        self.gpib_addr = kwargs.get("gpib_addr",self.gpib_addr)
        # SET GPIB address on the device
        self._send("++addr " + str(self.gpib_addr) + "\n")

    def _set_controller_mode(self,C_Mode=True):
        # set gpib_ethernet into controller mode (True) or in device mode (False)
        if C_Mode:
            # controller mode
            self._send("++mode 1")
        else:
            # device mode
            self._send("++mode 0")

    def _set_read_after_write(self, On=True):
        if On:
            # Turn on read-after-write
            self.sock.send("++auto 1\r\n".encode())
        else:
            # Turn off read-after-write to avoid "Query Unterminated" errors
            self.sock.send("++auto 0\r\n".encode())

    def _set_read_timeout(self,**kwargs):
        timeout = kwargs.get("timeout",self.timeout)
        if timeout > 3:
            timeout = 3
        self._send("++read_tmo_ms "+str(timeout*1000))

    def _set_EOI_assert(self,On=True): #
        # Assert EOI signal line with last byte to indicate end of data
        if On:
            self._send("++eoi 1\n")
        else:
            self._send("++eoi 0\n")

    def _set_GPIB_EOS(self,EOS='\n'): # end of signal/string
        EOSs={'\r\n':0,'\r':1,'\n':2,'':3}
        self.sock.send("++eos"+EOSs.get(EOS)+"\n".encode())

    def _set_GPIB_EOT(self,EOT=False):
        # send at EOI an EOT (end of transmission) character ?
        if EOT:
            self.sock.send("++eot_enable 1\n".encode())
        else:
            self.sock.send("++eot_enable 0\n".encode())
    def _set_GPIB_EOT_char(self,EOT_char=42):
        # set the EOT character
        self.sock.send("++eot_char"+EOT_char+"\n".encode())

    def _set_ifc(self):
        self._send("++ifc")

    # get the service request bit
    def _get_srq(self,**kwargs):
        cmd="++srq"
        bufflen=kwargs.get("bufflen",self.chunk_size)
        cmd=cmd.rstrip()
        cmd+=self.term_char

        self._send(cmd)
        #self._set_read()
        return self.sock.recv(bufflen)

    def _get_spoll(self):
        cmd="++spoll"
        cmd=cmd.rstrip()
        cmd+=self.term_char
        self._send(cmd)
        return self.sock.recv(self.chunk_size)

    def _get_status(self):
        cmd="++status 48"
        #cmd=cmd.rstrip()
        cmd+=self.term_char
        self._send(cmd)
        return self.sock.recv(self.chunk_size)

    def _set_reset(self):
        # Reset Device GPIB endpoint
        self._send("++rst")

    def _set_GPIB_dev_reset(self):
        # Reset Device GPIB endpoint
        self._send("*RST")

    def _get_idn(self):
        return self._send_recv("*IDN?")

    def _dump_internal_vars(self):
        print("timeout"+ str(self.timeout))
        print("chunk_size"+ str(self.chunk_size))
        print("values_format"+ str(self.values_format))
        print("term_char"+ str(self.term_char))
        print("send_end"+ str(self.send_end))
        print("delay"+ str(self.delay))
        print("lock"+str(self.lock))
        print("gpib_addr"+ str(self.gpib_addr))
        print("ip"+ str(self.ip))
        print("ethernet_port"+ str(self.ethernet_port))

    def CheckError(self):
        # check for device error
        self.sock.send("SYST:ERR?\n".encode())
        self.sock.send("++read eoi\n".encode())

        s = None
        try:
            s = self.sock.recv(100)
        except socket.timeout:
            print("socket timeout")
            s = ""
        except socket.error as e:
            pass
        print("Prologix CheckError():{}".format(s))



class errors():
    SPCM_ERROR_ORIGIN_MASK = (-2147483648)
    SPCM_ERROR_ORIGIN_LOCAL = 0x00000000
    SPCM_ERROR_ORIGIN_REMOTE = (-2147483648)
    ERR_OK = 0x0000
    ERR_INIT = 0x0001
    ERR_NR = 0x0002
    ERR_TYP = 0x0003
    ERR_FNCNOTSUPPORTED = 0x0004
    ERR_BRDREMAP = 0x0005
    ERR_KERNELVERSION = 0x0006
    ERR_HWDRVVERSION = 0x0007
    ERR_ADRRANGE = 0x0008
    ERR_INVALIDHANDLE = 0x0009
    ERR_BOARDNOTFOUND = 0x000A
    ERR_BOARDINUSE = 0x000B
    ERR_EXPHW64BITADR = 0x000C
    ERR_FWVERSION = 0x000D
    ERR_LASTERR = 0x0010
    ERR_ABORT = 0x0020
    ERR_BOARDLOCKED = 0x0030
    ERR_DEVICE_MAPPING = 0x0032
    ERR_NETWORKSETUP = 0x0040
    ERR_NETWORKTRANSFER = 0x0041
    ERR_FWPOWERCYCLE = 0x0042
    ERR_NETWORKTIMEOUT = 0x0043
    ERR_BUFFERSIZE = 0x0044
    ERR_RESTRICTEDACCESS = 0x0045
    ERR_INVALIDPARAM = 0x0046
    ERR_REG = 0x0100
    ERR_VALUE = 0x0101
    ERR_FEATURE = 0x0102
    ERR_SEQUENCE = 0x0103
    ERR_READABORT = 0x0104
    ERR_NOACCESS = 0x0105
    ERR_POWERDOWN = 0x0106
    ERR_TIMEOUT = 0x0107
    ERR_CALLTYPE = 0x0108
    ERR_EXCEEDSINT32 = 0x0109
    ERR_NOWRITEALLOWED = 0x010A
    ERR_SETUP = 0x010B
    ERR_CLOCKNOTLOCKED = 0x010C
    ERR_MEMINIT = 0x010D
    ERR_POWERSUPPLY = 0x010E
    ERR_ADCCOMMUNICATION = 0x010F
    ERR_CHANNEL = 0x0110
    ERR_NOTIFYSIZE = 0x0111
    ERR_RUNNING = 0x0120
    ERR_ADJUST = 0x0130
    ERR_PRETRIGGERLEN = 0x0140
    ERR_DIRMISMATCH = 0x0141
    ERR_POSTEXCDSEGMENT = 0x0142
    ERR_SEGMENTINMEM = 0x0143
    ERR_MULTIPLEPW = 0x0144
    ERR_NOCHANNELPWOR = 0x0145
    ERR_ANDORMASKOVRLAP = 0x0146
    ERR_ANDMASKEDGE = 0x0147
    ERR_ORMASKLEVEL = 0x0148
    ERR_EDGEPERMOD = 0x0149
    ERR_DOLEVELMINDIFF = 0x014A
    ERR_STARHUBENABLE = 0x014B
    ERR_PATPWSMALLEDGE = 0x014C
    ERR_NOPCI = 0x0200
    ERR_PCIVERSION = 0x0201
    ERR_PCINOBOARDS = 0x0202
    ERR_PCICHECKSUM = 0x0203
    ERR_DMALOCKED = 0x0204
    ERR_MEMALLOC = 0x0205
    ERR_EEPROMLOAD = 0x0206
    ERR_CARDNOSUPPORT = 0x0207
    ERR_CONFIGACCESS = 0x0208
    ERR_FIFOBUFOVERRUN = 0x0300
    ERR_FIFOHWOVERRUN = 0x0301
    ERR_FIFOFINISHED = 0x0302
    ERR_FIFOSETUP = 0x0309
    ERR_TIMESTAMP_SYNC = 0x0310
    ERR_STARHUB = 0x0320
    ERR_INTERNAL_ERROR = 0xFFFF





# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'plasma1_gui.ui'
#
# Created: Tue Jan 10 22:45:21 2017
#      by: PyQt4 UI code generator 4.11.3
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):
        return s

try:
    _encoding = QtGui.QApplication.UnicodeUTF8
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig)

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName(_fromUtf8("MainWindow"))
        MainWindow.setEnabled(True)
        MainWindow.resize(1292, 451)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Preferred, QtGui.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(MainWindow.sizePolicy().hasHeightForWidth())
        MainWindow.setSizePolicy(sizePolicy)
        self.centralwidget = QtGui.QWidget(MainWindow)
        self.centralwidget.setObjectName(_fromUtf8("centralwidget"))
        self.graph_MC = PlotWidget(self.centralwidget)
        self.graph_MC.setGeometry(QtCore.QRect(650, 10, 630, 350))
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.graph_MC.sizePolicy().hasHeightForWidth())
        self.graph_MC.setSizePolicy(sizePolicy)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        self.graph_MC.setBackgroundBrush(brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.NoBrush)
        self.graph_MC.setForegroundBrush(brush)
        self.graph_MC.setObjectName(_fromUtf8("graph_MC"))
        self.graph_LL = PlotWidget(self.centralwidget)
        self.graph_LL.setGeometry(QtCore.QRect(10, 10, 630, 350))
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.graph_LL.sizePolicy().hasHeightForWidth())
        self.graph_LL.setSizePolicy(sizePolicy)
        self.graph_LL.setFrameShadow(QtGui.QFrame.Sunken)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        self.graph_LL.setBackgroundBrush(brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.NoBrush)
        self.graph_LL.setForegroundBrush(brush)
        self.graph_LL.setObjectName(_fromUtf8("graph_LL"))
        self.label_LL = QtGui.QLabel(self.centralwidget)
        self.label_LL.setGeometry(QtCore.QRect(220, 380, 211, 31))
        font = QtGui.QFont()
        font.setFamily(_fromUtf8("LM Sans 10"))
        font.setPointSize(16)
        self.label_LL.setFont(font)
        self.label_LL.setTextFormat(QtCore.Qt.PlainText)
        self.label_LL.setObjectName(_fromUtf8("label_LL"))
        self.label_MC = QtGui.QLabel(self.centralwidget)
        self.label_MC.setGeometry(QtCore.QRect(860, 380, 231, 31))
        font = QtGui.QFont()
        font.setFamily(_fromUtf8("LM Sans 10"))
        font.setPointSize(16)
        self.label_MC.setFont(font)
        self.label_MC.setTextFormat(QtCore.Qt.PlainText)
        self.label_MC.setObjectName(_fromUtf8("label_MC"))
        self.cb_scale = QtGui.QComboBox(self.centralwidget)
        self.cb_scale.setGeometry(QtCore.QRect(620, 390, 74, 21))
        font = QtGui.QFont()
        font.setFamily(_fromUtf8("LM Sans 10"))
        font.setPointSize(11)
        self.cb_scale.setFont(font)
        self.cb_scale.setContextMenuPolicy(QtCore.Qt.DefaultContextMenu)
        self.cb_scale.setObjectName(_fromUtf8("cb_scale"))
        self.cb_scale.addItem(_fromUtf8(""))
        self.cb_scale.addItem(_fromUtf8(""))
        self.cb_scale.addItem(_fromUtf8(""))
        self.cb_scale.addItem(_fromUtf8(""))
        self.label = QtGui.QLabel(self.centralwidget)
        self.label.setGeometry(QtCore.QRect(560, 390, 71, 21))
        font = QtGui.QFont()
        font.setFamily(_fromUtf8("LM Sans 10"))
        font.setPointSize(11)
        self.label.setFont(font)
        self.label.setObjectName(_fromUtf8("label"))
        MainWindow.setCentralWidget(self.centralwidget)
        self.statusbar = QtGui.QStatusBar(MainWindow)
        self.statusbar.setEnabled(True)
        self.statusbar.setObjectName(_fromUtf8("statusbar"))
        MainWindow.setStatusBar(self.statusbar)
        self.actionQuit = QtGui.QAction(MainWindow)
        self.actionQuit.setObjectName(_fromUtf8("actionQuit"))
        self.actionStart = QtGui.QAction(MainWindow)
        self.actionStart.setObjectName(_fromUtf8("actionStart"))
        self.b_auto = QtGui.QAction(MainWindow)
        self.b_auto.setObjectName(_fromUtf8("b_auto"))
        self.b_24 = QtGui.QAction(MainWindow)
        self.b_24.setObjectName(_fromUtf8("b_24"))
        self.b_12 = QtGui.QAction(MainWindow)
        self.b_12.setObjectName(_fromUtf8("b_12"))
        self.b_6 = QtGui.QAction(MainWindow)
        self.b_6.setObjectName(_fromUtf8("b_6"))
        self.b_2 = QtGui.QAction(MainWindow)
        self.b_2.setObjectName(_fromUtf8("b_2"))
        self.b_1 = QtGui.QAction(MainWindow)
        self.b_1.setObjectName(_fromUtf8("b_1"))
        self.m_poll = QtGui.QAction(MainWindow)
        self.m_poll.setEnabled(False)
        self.m_poll.setObjectName(_fromUtf8("m_poll"))

        self.retranslateUi(MainWindow)
        self.cb_scale.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(_translate("MainWindow", "Plasma 1 monitor", None))
        self.label_LL.setText(_translate("MainWindow", "LL: 3.85e-8 mbar", None))
        self.label_MC.setText(_translate("MainWindow", "MC: 3.45e-9 mbar", None))
        self.cb_scale.setItemText(0, _translate("MainWindow", "today", None))
        self.cb_scale.setItemText(1, _translate("MainWindow", "2h", None))
        self.cb_scale.setItemText(2, _translate("MainWindow", "1h", None))
        self.cb_scale.setItemText(3, _translate("MainWindow", "10min", None))
        self.label.setText(_translate("MainWindow", "Scale:", None))
        self.actionQuit.setText(_translate("MainWindow", "Quit", None))
        self.actionStart.setText(_translate("MainWindow", "Start", None))
        self.b_auto.setText(_translate("MainWindow", "Autoscale", None))
        self.b_24.setText(_translate("MainWindow", "24h", None))
        self.b_12.setText(_translate("MainWindow", "12h", None))
        self.b_6.setText(_translate("MainWindow", "6h", None))
        self.b_2.setText(_translate("MainWindow", "2h", None))
        self.b_1.setText(_translate("MainWindow", "1h", None))
        self.m_poll.setText(_translate("MainWindow", "Measure all channels", None))

from pyqtgraph import PlotWidget



# transport.py  measurement class for IV like transport measurements
# Hannes Rotzinger, hannes.rotzinger@kit.edu 2010
# Micha Wildermuth, micha.wildermuth@kit.edu 2017

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import numpy as np
from scipy import signal
import logging
import time
import sys
import threading

import qkit
from qkit.storage import store as hdf
from qkit.gui.plot import plot as qviewkit
from qkit.gui.notebook.Progress_Bar import Progress_Bar
from qkit.measure.measurement_class import Measurement 
import qkit.measure.write_additional_files as waf


class transport(object):
    """
    This is a measurement class to perform transport measurements such as IV-characteristics by means of a current-voltage source meter.
    """
    
    def __init__(self, IV_Device):
        """
        Initializes the class for transport measurement such as IV characteristics.
        
        Parameters
        ----------
        IV_Device: qkit-instrument
            Current-voltage device that provides biasing and sensing.
        
        Returns
        -------
        None
        
        Examples
        --------
        >>> import qkit
        QKIT configuration initialized -> available as qkit.cfg[...]
        >>> qkit.cfg['run_id']='run_number'
        >>> qkit.cfg['user']='User'
        >>> qkit.start()
        Starting QKIT framework ... -> qkit.core.startup
        Loading module ... S10_logging.py
        Loading module ... S12_lockfile.py
        Loading module ... S14_setup_directories.py
        Loading module ... S20_check_for_updates.py
        Loading module ... S25_info_service.py
        Loading module ... S30_qkit_start.py
        Loading module ... S65_load_RI_service.py
        Loading module ... S70_load_visa.py
        Loading module ... S80_load_file_service.py
        Loading module ... S85_init_measurement.py
        Loading module ... S98_started.py
        Loading module ... S99_init_user.py
        Initialized the file info database (qkit.fid) in 0.000 seconds.
        
        >>> IVD = qkit.instruments.create('IVD', 'Keithley', address='TCPIP0::00.00.000.00::INSTR', reset=True)  # Keithley_2636A
        >>> from qkit.measure.transport import transport
        >>> tr = transport.transport(IV_Device=IVD)
        
        >>> import qkit.measure.samples_class as sc
        >>> sample = sc.Sample()
        >>> sample.name = 'sample_name'
        
        >>> tr.set_filename(filename='filename')
        >>> tr.set_expname(expname='expname')
        >>> tr.set_comment(comment='comment')
        >>> tr.set_sample(sample)
        >>> tr.set_dVdI(True)
        >>> tr.sweep.reset_sweeps()
        >>> tr.add_sweep_4quadrants(start=0e-9, stop=100e-9, step=100e-12, offset=0e-12)
        
        >>> tr.measure_XD()
        Measurement complete: d:\\data\XXXXXX_XD_sample_name\XXXXXX_XD_sample_name.h5
        Plots saved in d:\\data\XXXXXX_XD_sample_name\images
        """
        self._IVD = IV_Device
        # measurement object
        self._measurement_object = Measurement()
        self._measurement_object.measurement_type = 'transport'
        self._web_visible = True
        self._filename = None
        self._expname = None
        self._comment = None
        # measurement services
        self.progress_bar = True
        self.open_qviewkit = True
        self._qvk_process = False  # qviewkit process
        self._plot_comment = ''
        # measurement setting default values
        self.sweeps = self.sweep()  # calls sweep subclass
        # numerical derivation dV/dI (differential resistance)
        self._dVdI = False  # adds numerical derivation dV/dI as data series, views, ...
        self._numder_func = signal.savgol_filter  # function to calculate numerical derivative (default: Savitzky-Golay filter)
        self._numder_args = ()  # arguments for derivation function
        self._numder_kwargs = {'window_length': 15, 'polyorder': 3, 'deriv': 1}  # keyword arguments for derivation function
        self._average = None  # trace averaging
        self._view_xy = False
        # x and y data
        self._hdf_x = None
        self._hdf_y = None
        # xy, 2D & 3D scan variables
        self.set_log_function()
        self._x_name = ''
        self._y_name = ''
        self._scan_dim = None
        self._x_dt = 2e-3  # in s
        self._x_vec = [None]
        self._x_coordname = None
        self._x_set_obj = None
        self._x_unit = None
        self._tdy = 2e-3  # in s
        self._y_vec = [None]
        self._y_coordname = None
        self._y_set_obj = None
        self._y_unit = None
        self._landscape = False
        self._fit_func = None
        self._fit_name = None
        self._fit_unit = None
        self._fit_kwargs = None
    
    def set_sample(self, sample):
        """
        Adds sample object (e.g. with sample properties, measurement comments) to measurement object.
        
        Parameters
        ----------
        sample: qkit.measure.sample_class.Sample()
            Sample object to store information about the sample.
        
        Returns
        -------
        None
        
        Examples
        --------
        >>> import qkit.measure.samples_class as sc
        >>> sample = sc.Sample()
        >>> sample.name = 'sample_name'
        >>> tr.set_sample(sample)
        """
        self._measurement_object.sample = sample
        return
    
    def add_sweep_4quadrants(self, start, stop, step, offset=0, sleep=0):
        """
        Adds a four quadrants sweep series with the pattern 
            0th: (+<start> --> +<stop>,  <step>) + <offset>
            1st: (+<stop>  --> +<start>, <step>) + <offset>
            2nd: (+<start> --> -<stop>,  <step>) + <offset>
            3rd: (-<stop>  --> +<start>, <step>) + <offset>
            time.sleep(<sleep>)
        
        Parameters
        ----------
        start: float
            Start value of sweep.
        stop: float
            Stop value of sweep.
        step: float
            Step value of sweep.
        offset: float, optional
            Offset value by which <start> and <stop> are shifted. Default is 0.
        sleep: float
            Sleep time after whole sweep. Default is 0.
        
        Returns
        -------
        None
        """
        self.sweeps.add_sweep(start=start+offset, stop=+stop+offset, step=step, sleep=sleep)
        self.sweeps.add_sweep(start=+stop+offset, stop=start+offset, step=step, sleep=sleep)
        self.sweeps.add_sweep(start=start+offset, stop=-stop+offset, step=step, sleep=sleep)
        self.sweeps.add_sweep(start=-stop+offset, stop=start+offset, step=step, sleep=sleep)
        return
    
    def add_sweep_halfswing(self, amplitude, step, offset=0, sleep=0):
        """
        Adds a halfswing sweep series with the pattern 
            0th: (+<amplitude> --> -<amplitude, <step>) + <offset>
            1st: (-<amplitude> --> +<amplitude, <step>) + <offset>
            time.sleep(<sleep>)
        
        Parameters
        ----------
        amplitude: float
            Amplitude value of sweep.
        step: float
            Step value of sweep.
        offset: float
            Offset value by which <start> and <stop> are shifted. Default is 0.
        sleep: float
            Sleep time after whole sweep. Default is 0.
        
        Returns
        -------
        None
        """
        self.sweeps.add_sweep(start=+amplitude+offset, stop=-amplitude+offset, step=step, sleep=sleep)
        self.sweeps.add_sweep(start=-amplitude+offset, stop=+amplitude+offset, step=step, sleep=sleep)
        return
    
    def set_dVdI(self, status, func=None, *args, **kwargs):
        """
        Sets the internal dVdI parameter, to decide weather the differential resistance (dV/dI) is calculated or not.
        
        Parameters
        ----------
        status: bool
            Status if numerical derivative is calculated.
        func: function
            Function to calculate numerical derivative, e.g. scipy.signal.savgol_filter (default), numpy.gradient, ...
        *args: array_likes, optional
            Arguments for derivation function.
        **kwargs: dictionary_likes, optional
            Keyword arguments for derivation function.
        
        Returns
        -------
        None
        
        Examples
        --------
        Savitzky-Golay filter
        >>> tr.set_dVdI(True, func=signal.savgol_filter, window_length=15, polyorder=3, deriv=1)
        
        Gradient as difference quotient
        >>> tr.set_dVdI(True, func=np.gradient)
        """
        self._dVdI = status
        if func is not None:
            self._numder_func = func
            self._numder_args = args
            self._numder_kwargs = kwargs
        return
    
    def get_dVdI(self):
        """
        Gets the internal dVdI parameter, to decide weather the differential resistance (dV/dI) is calculated or not.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        status: bool
            Status if numerical derivative is calculated.
        func: function
            Function to calculate numerical derivative, e.g. scipy.signal.savgol_filter (default), numpy.gradient, ...
        *args: array_likes
            Arguments for derivation function.
        **kwargs: dictionary_likes
            Keyword arguments for derivation function.
        """
        return self._dVdI, self._numder_func, self._numder_args, self._numder_kwargs
    
    def _numerical_derivative(self, x, y):
        """
        Calculates numerical derivative dy/dx by means of set function <self._numder_func>, arguments <self._numder_args> and keyword arguments <self._numder_kwargs>.
    
        Parameters
        ----------
        x: array_likes
            An N-dimensional array containing x-values for numerical derivative.
        y: array_likes
            An N-dimensional array containing y-values for numerical derivative.
    
        Returns
        -------
        dydx: numpy.array
            An N-dimensional array containing numerical derivative dy/dx.
        """
        # TODO: catch error, if len(dataset) < window_length in case of SavGol filter
        try:
            return self._numder_func(y, *self._numder_args, **self._numder_kwargs)/self._numder_func(x, *self._numder_args, **self._numder_kwargs)
        except Exception as e:
            logging.warning("Can't calculate numerical derivative, possibly insufficient data points. %s", e)
            return np.zeros(len(y))*np.nan
    
    def set_x_dt(self, x_dt):
        """
        Sets sleep time between x-iterations in 2D and 3D scans.
        
        Parameters
        ----------
        x_dt: float
            The sleep time between x-iterations.
        
        Returns
        -------
        None
        """
        self._x_dt = x_dt
        return
    
    def get_x_dt(self):
        """
        Gets sleep time between x-iterations in 2D and 3D scans.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        x_dt: float
            The sleep time between x-iterations
        """
        return self._x_dt
    
    def set_x_parameters(self, x_vec, x_coordname, x_set_obj, x_unit='', x_dt=None):
        """
        Sets x-parameters for 2D and 3D scan.
        In a 3D measurement, the x-parameters will be the "outer" sweep meaning for every x value all y values are swept and for each (x,y) value the bias is swept according to the set sweep parameters.
        
        Parameters
        ----------
        x_vec: array_likes
            An N-dimensional array that contains the sweep values.
        x_coordname: string
            The coordinate name to be created as data series in the .h5 file.
        x_set_obj: obj
            An callable object to execute with x_vec-values.
        x_unit: string, optional
            The unit name to be used in data series in the .h5 file.
        x_dt: float, optional
            The sleep time between x-iterations.
        
        Returns
        -------
        None
        """
        # x-vec
        if np.iterable(x_vec):
            try:
                self._x_vec = np.array(x_vec, dtype=float)
            except Exception as e:
                raise type(e)('{!s}: Cannot set {!s} as x-vector'.format(__name__, x_vec, e))
        else:
            raise TypeError('{:s}: Cannot set {!s} as x-vector: iterable object needed'.format(__name__, x_vec))
        # x-coordname
        if type(x_coordname) is str:
            self._x_coordname = x_coordname
        else:
            raise ValueError('{:s}: Cannot set {!s} as x-coordname: string needed'.format(__name__, x_coordname))
        # x-set-object
        if callable(x_set_obj):
            self._x_set_obj = x_set_obj
        else:
            raise ValueError('{:s}: Cannot set {!s} as x-set-object: callable object needed'.format(__name__, x_set_obj))
        # x-unit
        if type(x_unit) is str:
            self._x_unit = x_unit
        else:
            raise ValueError('{:s}: Cannot set {!s} as x-unit: string needed'.format(__name__, x_unit))
        # x dt
        if x_dt is not None:
            self._x_dt = x_dt
        if self._landscape:
            self._lsc_vec = self._lsc_func(np.array(self._x_vec), *self._lsc_args)
        return
    
    def set_tdy(self, x_dt):
        """
        Sets sleep time between y-iterations in 3D scans.
        
        Parameters
        ----------
        x_dt: float
            The sleep time between y-iterations.
        
        Returns
        -------
        None
        """
        self._tdy = x_dt
        return
    
    def get_tdy(self):
        """
        Gets sleep time between y-iterations in 3D scans.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        x_dt: float
            The sleep time between y-iterations.
        """
        return self._tdy
    
    def set_y_parameters(self, y_vec, y_coordname, y_set_obj, y_unit='', y_dt=None):
        """
        Sets x-parameters for 3D scan, where y-parameters will be the "inner" sweep meaning for every x value all y values are swept and for each (x,y) value the bias is swept according to the set sweep parameters.
        
        Parameters
        ----------
        y_vec: array_likes
            An N-dimensional array that contains the sweep values.
        y_coordname: string
            The coordinate name to be created as data series in the .h5 file.
        y_set_obj: obj
            An callable object to execute with y_vec-values.
        y_unit: string, optional
            The unit name to be used in data series in the .h5 file.
        y_dt: float, optional
            The sleep time between y-iterations.
        
        Returns
        -------
        None
        """
        # y-vec
        if np.iterable(y_vec):
            try:
                self._y_vec = np.array(y_vec, dtype=float)
            except Exception as e:
                raise type(e)('{!s}: Cannot set {!s} as y-vector'.format(__name__, y_vec, e))
        else:
            raise TypeError('{:s}: Cannot set {!s} as y-vector: iterable object needed'.format(__name__, y_vec))
        # y-coordname
        if type(y_coordname) is str:
            self._y_coordname = y_coordname
        else:
            raise ValueError('{:s}: Cannot set {!s} as y-coordname: string needed'.format(__name__, y_coordname))
        # y-set-object
        if callable(y_set_obj):
            self._y_set_obj = y_set_obj
        else:
            raise ValueError('{:s}: Cannot set {!s} as y-set-object: callable object needed'.format(__name__, y_set_obj))
        # y-unit
        if type(y_unit) is str:
            self._y_unit = y_unit
        else:
            raise ValueError('{:s}: Cannot set {!s} as y-unit: string needed'.format(__name__, y_unit))
        # y dt
        if y_dt is not None:
            self._y_dt = y_dt
        return
    
    def set_landscape(self, func, args, mirror=True):
        """
        Sets parameters for landscapes scans in case of 2D and 3D scans to fasten up the measurement.
        The values of the given envelop function limit the sweep bounds for each x-value of 2D or 3D scans. The overall sweep bounds can merely be decreased, so that only envelop values <= sweep bounds affect them.
        
        Parameters
        ----------
        func: function
            An envelop function that limits bias values.
        args: tuple
            The arguments for the envelop function.
        mirror: bool, optional
            This determines if the envelop function is mirrored at the x-axis. Default is True
        
        Returns
        -------
        None

        Examples
        --------
        >>> tr.add_sweep_halfswing(1e-6, 10e-9)
        >>> def x_function(i):
                set_x_value(i)
                return
        >>> tr.set_x_parameters(x_vec = np.linspace(-10, 10, 21),
                                x_coordname = 'x_name',
                                x_set_obj = x_function,
                                x_unit = 'x_unit')
        >>> def lsc_func(x, a0, a1, x0):
                y = np.ones(shape = len(x)) * a1
                y[np.abs(x) > x0] = a0
                return y
        >>> tr.set_landscape(lsc_func, (500e-9, 1e-6, 5))
        """
        # TODO: possibility for landscape scans in both x and y direction
        self._landscape = True
        self._lsc_func = func
        self._lsc_args = args
        self._lsc_vec = func(np.array(self._x_vec), *args)
        self._lsc_mirror = mirror
        return
    
    def reset_landscape(self):
        self._landscape = False
        self._lsc_vec = None
        self._lsc_mirror = False
        return
    
    def set_xy_parameters(self, x_name, x_func, x_vec, x_unit, y_name, y_func, y_unit, x_kwargs={}, y_kwargs={}, x_dt=1e-3):
        """
        Set x- and y-parameters for measure_xy(), where y-parameters can be a list in order to record various quantities.
        
        Parameters
        ----------
        x_name: str
            The name of the x-parameter to be created as data series in the .h5 file.
        x_func: function
            A function that returns x-values.
        x_vec: array_likes of floats
            An array of x-values at which data should be taken.
        x_unit: str
            The unit of the x-parameter to be used in data series in the .h5 file.
        x_kwargs: dict
            Keyword arguments forwarded to the x-function <x_func>.
        y_name: array_likes of strings
            Names of y-parameters to be created as data series in the .h5 file.
        y_func: function
            A function that returns y-values.
        y_unit: array_likes of strings
            Units of y-parameters to be used in data series in the .h5 file.
        y_kwargs: dict
            Keyword arguments forwarded to the y-function <y_func>.
        x_dt: float, optional
            The sleep time between queries of x-values. Default is 1e-3.
        
        Returns
        -------
        None
        """
        # x-name
        if type(x_name) is str:
            self._x_name = x_name
        else:
            raise ValueError('{:s}: Cannot set {!s} as x-name: string needed'.format(__name__, x_name))
        # x-func
        if callable(x_func):
            self._x_func = x_func
        else:
            raise ValueError('{:s}: Cannot set {!s} as x-function: callable object needed'.format(__name__, x_func))
        # x-vec
        if np.iterable(x_vec):
            self._x_vec = x_vec
        else:
            raise ValueError('{:s}: Cannot set {!s} as x-vector: iterable object needed'.format(__name__, x_vec))
        # x-unit
        if type(x_unit) is str:
            self._x_unit = x_unit
        else:
            raise ValueError('{:s}: Cannot set {!s} as x-unit: string needed'.format(__name__, x_unit))
        # x-kwargs
        if type(x_kwargs) is dict:
            self._x_kwargs = x_kwargs
        else:
            raise ValueError('{:s}: Cannot set {!s} as x-kwargs: dict needed'.format(__name__, x_kwargs))
        # y-name
        if type(y_name) is str:
            self._y_name = [y_name]
        elif np.iterable(y_name):
            for name in y_name:
                if type(name) is not str:
                    raise ValueError('{:s}: Cannot set {!s} as y-name: string needed'.format(__name__, name))
            self._y_name = y_name
        else:
            raise ValueError('{:s}: Cannot set {!s} as y-name: string of iterable object of strings needed'.format(__name__, y_name))
        # y-func
        if callable(y_func):
            self._y_func = [y_func]
        elif np.iterable(y_func):
            for func in y_func:
                if not callable(func):
                    raise ValueError('{:s}: Cannot set {!s} as y-function: callable object needed'.format(__name__, func))
            self._y_func = y_func
        else:
            raise ValueError('{:s}: Cannot set {!s} as y-function: callable object of iterable object of callable objects needed'.format(__name__, y_func))
        # y-unit
        if type(y_unit) is str:
            self._y_unit = [y_unit]
        elif np.iterable(y_unit):
            for unit in y_unit:
                if type(unit) is not str:
                    raise ValueError('{:s}: Cannot set {!s} as y-unit: string needed'.format(__name__, unit))
            self._y_unit = y_unit
        else:
            raise ValueError('{:s}: Cannot set {!s} as y-unit: string of iterable object of strings needed'.format(__name__, y_unit))
        # y-kwargs
        if type(y_kwargs) is dict:
            self._y_kwargs = [y_kwargs]
        elif np.iterable(y_kwargs):
            for kwargs in y_kwargs:
                if type(kwargs) is not dict:
                    raise ValueError('{:s}: Cannot set {!s} as y-kwargs: dict needed'.format(__name__, kwargs))
            self._y_kwargs = y_kwargs
        else:
            raise ValueError('{:s}: Cannot set {!s} as y-kwargs: dict of iterable object of dicts needed'.format(__name__, y_kwargs))
        # x-dt
        self._x_dt = x_dt
        return
    
    def set_log_function(self, func=None, name=None, unit=None, dtype='f'):
        """
        Saves desired values obtained by a function <func> in the .h5-file as a value vector with name <name>, unit <unit> and in data format <f>.
        The function (object) can be passed to the measurement loop which is executed before every x iteration
        but after executing the x_object setter in 2D measurements and before every line (but after setting 
        the x value) in 3D measurements.
        The return value of the function of type float or similar is stored in a value vector in the h5 file.
        
        Parameters
        ----------
        func: array_likes of callable objects
            A callable object that returns the value to be saved.
        name: array_likes of strings
            Names of logging parameter appearing in h5 file. Default is 'log_param'.
        unit: array_likes of strings
            Units of logging parameter. Default is ''.
        dtype: array_likes of dtypes
            h5 data type to be used in the data file. Default is 'f' (float64).
        
        Returns
        -------
        None
        """
        # TODO: dtype = float instead of 'f'
        # log-function
        if callable(func):
            func = [func]
        elif func is None:
            func = [None]
        elif np.iterable(func):
            for fun in func:
                if not callable(fun):
                    raise ValueError('{:s}: Cannot set {!s} as y-function: callable object needed'.format(__name__, fun))
        else:
            raise ValueError('{:s}: Cannot set {!s} as log-function: callable object of iterable object of callable objects needed'.format(__name__, func))
        self.log_function = func
        # log-name
        if name is None:
            try:
                name = ['log_param']*len(func)
            except Exception:
                name = [None]
        elif type(name) is str:
            name = [name]*len(func)
        elif np.iterable(name):
            for _name in name:
                if type(_name) is not str:
                    raise ValueError('{:s}: Cannot set {!s} as log-name: string needed'.format(__name__, _name))
        else:
            raise ValueError('{:s}: Cannot set {!s} as log-name: string of iterable object of strings needed'.format(__name__, name))
        self.log_name = name
        # log-unit
        if unit is None:
            try:
                unit = ['log_unit']*len(func)
            except Exception:
                unit = [None]
        elif type(unit) is str:
            unit = [unit]*len(func)
        elif np.iterable(unit):
            for _unit in unit:
                if type(_unit) is not str:
                    raise ValueError('{:s}: Cannot set {!s} as log-unit: string needed'.format(__name__, _unit))
        else:
            raise ValueError('{:s}: Cannot set {!s} as log-unit: string of iterable object of strings needed'.format(__name__, unit))
        self.log_unit = unit
        # log-dtype
        if dtype is None:
            try:
                dtype = [float]*len(func)
            except Exception:
                dtype = [None]
        elif type(dtype) is type:
            dtype = [dtype]*len(func)
        elif np.iterable(dtype):
            for _dtype in dtype:
                if type(_dtype) is not str:
                    raise ValueError('{:s}: Cannot set {!s} as log-dtype: string needed'.format(__name__, _dtype))
        else:
            raise ValueError('{:s}: Cannot set {!s} as log-dtype: string of iterable object of strings needed'.format(__name__, dtype))
        self.log_dtype = dtype
        return
    
    def get_log_function(self):
        """
        Gets the current log_function settings.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        func: array_likes of callable objects
            A callable object that returns the value to be saved.
        name: array_likes of strings
            Names of logging parameter appearing in h5 file. Default is 'log_param'.
        unit: array_likes of strings
            Units of logging parameter. Default is ''.
        dtype: array_likes of dtypes
            h5 data type to be used in the data file. Default is float (float64).
        """
        return [f.__name__ for f in self.log_function], self.log_name, self.log_unit, self.log_dtype
    
    def reset_log_function(self):
        """
        Resets all log_function settings.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        self.log_function = []
        self.log_name = []
        self.log_unit = []
        self.log_dtype = []
        self._hdf_log = []
        return

    def set_fit_IV(self, func, name, unit='', **kwargs):
        """
        Sets fit-parameters for an in situ IV fit.
        This function is called for each trace with ints kexword arguments

        Parameters
        ----------
        func: obj
            An callable object to execute for each trace.
        name: string
            The coordinate name to be created as data series in the .h5 file.
        unit: string, optional
            The unit name to be used in data series in the .h5 file.
        **kwargs: dictionary_likes, optional
            Keyword arguments for fit function.

        Returns
        -------
        None

        Examples
        --------
        >>> from qkit.analysis.IV_curve import IV_curve as IVC
        >>> ivc = IVC()

        Ic fit via threshold voltage
        >>> tr.set_fit_IV(func=ivc.get_Ic_threshold, name='I_c', unit='A', offset=0)

        Ic fit via peaks in the numerical derivative dV/dI
        >>> tr.set_fit_IV(func=ivc.get_Ic_deriv, name='I_c', unit='A', prominence=10)

        Ic fit via peaks in the smoothed derivation in the frequency domain
        >>> tr.set_fit_IV(func=ivc.get_Ic_dft, name='I_c', unit='A', prominence=1e-2)
        """
        self._fit_func = func
        self._fit_name = name
        self._fit_unit = unit
        self._fit_kwargs = kwargs

    def reset_fit_IV(self):
        """
        Resets all fit_function settings.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        self._fit_func = None
        self._fit_name = None
        self._fit_unit = None
        self._fit_kwargs = None

    def set_filename(self, filename):
        """
        Sets filename of current measurement to <filename>.
        
        Parameters
        ----------
        filename: str
            The file name used as suffix of uuid.
        
        Returns
        -------
        None
        """
        self._filename = filename
        return
    
    def get_filename(self):
        """
        Gets filename of current measurement.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        filename: str
            The file name used as suffix of uuid.
        """
        return self._filename
    
    def reset_filename(self):
        """
        Resets filename of current measurement to None.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        self._filename = None
        return
    
    def set_expname(self, expname):
        """
        Sets experiment name of current measurement to <expname>.
        
        Parameters
        ----------
        expname: str
            The experiment name used as suffix of uuid and <filename>.
        
        Returns
        -------
        None
        """
        self._expname = expname
        return
    
    def get_expname(self):
        """
        Gets experiment name of current measurement
        
        Parameters
        ----------
        None
        
        Returns
        -------
        expname: str
            The experiment name used as suffix of uuid and <filename>.
        """
        return self._expname
    
    def reset_expname(self):
        """
        Resets experiment name of current measurement to None.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        self._expname = None
        return
    
    def set_comment(self, comment):
        """
        Sets comment that is added to the .h5 file to <comment>
        
        Parameters
        ----------
        comment: str
            The comment added to data in .h5 file.
        
        Returns
        -------
        None
        """
        self._comment = comment
        return
    
    def get_comment(self):
        """
        Gets comment that is added to the .h5 file
        
        Parameters
        ----------
        None
        
        Returns
        -------
        comment: str
            The comment added to data in .h5 file.
        """
        return self._comment
    
    def reset_comment(self):
        """
        Resets comment that is added to the .h5 file to None.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        self._comment = None
        return
    
    def set_average(self, avg):
        """
        Sets trace average parameter.
        
        Parameters
        ----------
        avg: int
            Number of averages of whole traces. Must be None (off) or natural numbers.
        
        Returns
        -------
        None
        """
        self._average = avg
        return
    
    def get_average(self):
        """
        Sets trace average parameter.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        avg: int
            Number of averages of whole traces. Is None (off) or a natural number.
        """
        return self._average
    
    def set_view_xy(self, view):
        """
        Sets views that combine different data series of xy-measurement.
        
        Parameters
        ----------
        view: bool or array_likes of integers
            Parameter that determines if additional views of different y-parameters are added. Must be boolean or an array containing tuples of two natural numbers, where y1 is used x-axis and y2 as y-axis.
        
        Returns
        -------
        None
        """
        self._view_xy = view
        return
    
    def get_view_xy(self):
        """
        Gets views that combine different data series of xy-measurement.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        view: bool or array_likes of integers
            Parameter that determines if additional views of different y-parameters are added. Is boolean or an array containing tuples of two natural numbers, where y1 is used x-axis and y2 as y-axis.
        """
        return self._view_xy
    
    def measure_xy(self):
        """
        Measures single data points, e.g. current or voltage and iterating all parameters x_vec, e.g. time.
        Every single data point is taken with the current IV Device settings.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        self._scan_dim = 0
        ''' measurement object '''
        self._measurement_object.measurement_func = sys._getframe().f_code.co_name
        ''' measurement '''
        self._measure()
        return
    
    def measure_1D(self):
        """
        Measures a 1 dimensional set of IV curves while sweeping the bias according to the set sweep parameters. 
        Every single data point is taken with the current IV Device settings.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        self._scan_dim = 1
        ''' measurement object '''
        self._measurement_object.measurement_func = sys._getframe().f_code.co_name
        ''' measurement '''
        self._measure()
        return
    
    def measure_2D(self):
        """
        Measures a 2 dimensional set of IV curves while sweeping the bias according to the set sweep parameters and iterating all parameters x_vec in x_obj. 
        Every single data point is taken with the current IV Device settings.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        if self._x_set_obj is None:
            logging.error('{:s}: axes parameters not properly set'.format(__name__))
            raise TypeError('{:s}: axes parameters not properly set'.format(__name__))
        
        self._scan_dim = 2        
        ''' measurement object '''
        self._measurement_object.measurement_func = sys._getframe().f_code.co_name
        ''' measurement '''
        self._measure()
        return
    
    def measure_3D(self):
        """
        Measures a 3 dimensional set of IV curves while sweeping the bias according to the set sweep parameters and iterating all parameters x_vec in x_obj and all parameters y_vec in y_obj. The sweep over y_obj is the inner loop, for every value x_vec[i] all values y_vec are measured.
        Every single data point is taken with the current IV Device settings.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        if self._x_set_obj is None or self._y_set_obj is None:
            logging.error('{:s}: axes parameters not properly set'.format(__name__))
            raise TypeError('{:s}: axes parameters not properly set'.format(__name__))
        self._scan_dim = 3
        ''' measurement object '''
        self._measurement_object.measurement_func = sys._getframe().f_code.co_name
        ''' measurement '''
        self._measure()
        return
    
    def _measure(self):
        """
        Creates output files, measures according to IVD and sweep settings, stores data and shows them in the qviewkit
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        def _pass(arg):
            """ dummy function that just passes, used for x, y iteration in case of 1D and 2D scan """
            pass
        
        ''' axis labels '''
        _axis = {0: (self._x_name, '', ''),
                 1: ('voltage', '', ''),
                 2: (self._x_coordname, 'current', ''),  # or self._IV_modes[self._bias] for y
                 3: (self._x_coordname, self._y_coordname, 'current')}
        self._measurement_object.x_axis, self._measurement_object.y_axis, self._measurement_object.z_axis = _axis[self._scan_dim]
        ''' prepare IV device '''
        self._prepare_measurement_IVD()
        ''' prepare data storage '''
        self._prepare_measurement_file()
        ''' prepare progress bar '''
        self._prepare_progress_bar()
        ''' opens qviewkit to plot measurement '''
        if self.open_qviewkit:
            if self._scan_dim == 0:
                datasets = []
            else:
                datasets = ['views/IV']
                if self._fit_name:
                    for i in range(self.sweeps.get_nos()):
                        datasets.append('analysis/{:s}_{:d}'.format(self._fit_name, i))
                elif self._scan_dim > 1:
                    for i in range(self.sweeps.get_nos()):
                        datasets.append('{:s}_{:d}'.format(self._IV_modes[not self._bias].lower(), i))
            self._qvk_process = qviewkit.plot(self._data_file.get_filepath(), datasets=datasets)  # opens IV-view by default
        ''' measurement '''
        sys.stdout.flush()
        qkit.flow.start()
        try:
            if self._scan_dim == 0:  # single data points
                for x_val in self._x_vec:
                    self._hdf_x.add(x_val)
                    self._x_func(x_val, **self._x_kwargs)
                    for func, kwargs, lst in zip(self._y_func, self._y_kwargs, self._hdf_y):
                        lst.add(func(**kwargs))
                    # iterate progress bar
                    if self.progress_bar:
                        self._pb.iterate()
                    time.sleep(self._x_dt)
            elif self._scan_dim in [1, 2, 3]:  # IV curve
                _rst_log_hdf_appnd = False  # variable to save points of log-function in 2D-matrix
                self._rst_fit_hdf_appnd = False
                for self.ix, (x, x_func) in enumerate([(None, _pass)] if self._scan_dim < 2 else [(x, self._x_set_obj) for x in self._x_vec]):  # loop: x_obj with parameters from x_vec if 2D or 3D else pass(None)
                    x_func(x)
                    time.sleep(self._x_dt)
                    for self.iy, (y, y_func) in enumerate([(None, _pass)] if self._scan_dim < 3 else [(y, self._y_set_obj) for y in self._y_vec]):  # loop: y_obj with parameters from y_vec if 3D else pass(None)
                        y_func(y)
                        time.sleep(self._tdy)
                        # log function
                        if self.log_function != [None]:
                            for j, f in enumerate(self.log_function):
                                if self._scan_dim == 1:
                                    self._data_log[j] = np.array([float(f())])  # np.asarray(f(), dtype=float)
                                    self._hdf_log[j].append(self._data_log[j])
                                elif self._scan_dim == 2:
                                    self._data_log[j][self.ix] = float(f())
                                    self._hdf_log[j].append(self._data_log[j], reset=True)
                                elif self._scan_dim == 3:
                                    self._data_log[j][self.ix, self.iy] = float(f())
                                    self._hdf_log[j].append(self._data_log[j][self.ix], reset=_rst_log_hdf_appnd)
                            if self._scan_dim == 3: # reset needs to be updated for all log-functions simultaneously and thus outside of the loop 
                                _rst_log_hdf_appnd = not bool(self.iy+1 == len(self._y_vec))
                        # iterate sweeps and take data
                        self._get_sweepdata()
                    # filling of value-box by storing data in the next 2d structure after every y-loop
                    if self._scan_dim is 3:
                        for lst in [val for k, val in enumerate([self._hdf_I, self._hdf_V, self._hdf_dVdI]) if k < 2+int(self._dVdI)]:
                            for val in range(self.sweeps.get_nos()):
                                lst[val].next_matrix()
        finally:
            ''' end measurement '''
            qkit.flow.end()
            t = threading.Thread(target=qviewkit.save_plots, args=[self._data_file.get_filepath(), self._plot_comment])
            t.start()
            self._data_file.close_file()
            waf.close_log_file(self._log_file)
            self._set_IVD_status(False)
            print('Measurement complete: {:s}'.format(self._data_file.get_filepath()))
        return
    
    def _prepare_measurement_IVD(self):
        """
        All the relevant settings from the IVD are updated and called.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        self._sweep_mode = self._IVD.get_sweep_mode()  # 0 (VV-mode) | 1 (IV-mode) | 2 (VI-mode)
        self._bias = self._IVD.get_sweep_bias()  # 0 (current bias) | 1 (voltage bias)
        self._IV_modes = {0: 'I', 1: 'V'}
        self._IV_units = {0: 'A', 1: 'V'}
        self._set_IVD_status(True)
        return
    
    def _set_IVD_status(self, status):
        """
        Sets the output status of the used IVD (of channel <**kwargs>) to <status>.
        
        Parameters
        ----------
        status: bool
            Output status of used IV-device.
        
        Returns
        -------
        None
        """
        for channel in self._IVD.get_sweep_channels():
            self._IVD.set_status(status=status, channel=channel)
        return
    
    def _prepare_measurement_file(self):
        """
        Creates one file each for data (.h5) with distinct dataset structures for each measurement dimension, settings (.set), logging (.log) and measurement (.measurement).
        At this point all measurement parameters are known and put in the output files.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        ''' create files '''
        # data.h5 file
        self._data_file = hdf.Data(name='_'.join(list(filter(None, ('xy' if self._scan_dim is 0 else '{:d}D_IV_curve'.format(self._scan_dim), self._filename, self._expname)))), mode='a')
        # settings.set file
        self._hdf_settings = self._data_file.add_textlist('settings')
        self._hdf_settings.append(waf.get_instrument_settings(self._data_file.get_filepath()))
        # logging.log file
        self._log_file = waf.open_log_file(self._data_file.get_filepath())
        ''' measurement object, sample object '''
        self._measurement_object.uuid = self._data_file._uuid
        self._measurement_object.hdf_relpath = self._data_file._relpath
        self._measurement_object.instruments = qkit.instruments.get_instrument_names()  # qkit.instruments.get_instruments() #
        if self._measurement_object.sample is not None:
            self._measurement_object.sample.sweeps = self.sweeps.get_sweeps()
            self._measurement_object.sample.average = self._average
        if self._average is not None:
            self._measurement_object.average = self._average
        self._measurement_object.save()
        self._hdf_mo = self._data_file.add_textlist('measurement')
        self._hdf_mo.append(self._measurement_object.get_JSON())
        ''' data variables '''
        self._hdf_bias = []
        self._hdf_I = []
        self._hdf_V = []
        self._hdf_dVdI = []
        self._hdf_fit = []
        self._data_fit = []
        if self._scan_dim == 0:
            ''' xy '''
            # add data variables
            self._hdf_x = self._data_file.add_coordinate(self._x_name,
                                                         unit=self._x_unit)
            self._hdf_y = [self._data_file.add_value_vector(self._y_name[i],
                                                            x=self._hdf_x,
                                                            unit=self._y_unit[i],
                                                            save_timestamp=False)
                           for i,_ in enumerate(self._y_func)]
            # add views
            if self._view_xy:
                if self._view_xy is True:
                    self._view_xy = [(x, y) for i, x in enumerate(range(len(self._hdf_y)))
                                     for y, _ in enumerate(self._hdf_y)[i+1::]]
                for view in self._view_xy:
                    self._data_file.add_view('{:s}_vs_{:s}'.format(*np.array(self._y_name)[np.array(view[::-1])]),
                                             x=self._hdf_y[view[0]],
                                             y=self._hdf_y[view[1]])
        elif self._scan_dim == 1:
            ''' 1D scan '''
            # add data variables
            self.sweeps.create_iterator()
            for i in range(self.sweeps.get_nos()):
                self._hdf_bias.append(self._data_file.add_coordinate('{:s}_b_{!s}'.format(self._IV_modes[self._bias], i),
                                                                     unit=self._IV_units[self._bias]))
                self._hdf_bias[i].add(self._get_bias_values(sweep=self.sweeps.get_sweep()))
                self._hdf_I.append(self._data_file.add_value_vector('I_{!s}'.format(i),
                                                                    x=self._hdf_bias[i],
                                                                    unit='A',
                                                                    save_timestamp=False))
                self._hdf_V.append(self._data_file.add_value_vector('V_{!s}'.format(i),
                                                                    x=self._hdf_bias[i],
                                                                    unit='V',
                                                                    save_timestamp=False))
                if self._dVdI:
                    self._hdf_dVdI.append(self._data_file.add_value_vector('dVdI_{!s}'.format(i),
                                                                           x=self._hdf_bias[i],
                                                                           unit='V/A',
                                                                           save_timestamp=False,
                                                                           folder='analysis',
                                                                           comment=self._get_numder_comment(self._hdf_V[i].name)+
                                                                                   '/'+self._get_numder_comment(self._hdf_I[i].name)))
                if self._fit_func:
                    self._hdf_fit.append(self._data_file.add_value_vector('{:s}_{:d}'.format(self._fit_name, i),
                                                                          x=np.zeros(1),
                                                                          unit=self._fit_unit,
                                                                          save_timestamp=False,
                                                                          folder='analysis',
                                                                          comment=self._get_fit_comment(i)))
                    self._data_fit.append(np.nan)
            # log-function
            self._add_log_value_vector()
            # add views
            self._add_views()
        elif self._scan_dim == 2:
            ''' 2D scan '''
            self._hdf_x = self._data_file.add_coordinate(self._x_coordname, unit=self._x_unit)
            self._hdf_x.add(self._x_vec)
            # add data variables
            self.sweeps.create_iterator()
            for i in range(self.sweeps.get_nos()):
                self._hdf_bias.append(self._data_file.add_coordinate('{:s}_b_{!s}'.format(self._IV_modes[self._bias], i),
                                                                     unit=self._IV_units[self._bias]))
                self._hdf_bias[i].add(self._get_bias_values(sweep=self.sweeps.get_sweep()))
                self._hdf_I.append(self._data_file.add_value_matrix('I_{!s}'.format(i),
                                                                    x=self._hdf_x,
                                                                    y=self._hdf_bias[i],
                                                                    unit='A',
                                                                    save_timestamp=False))
                self._hdf_V.append(self._data_file.add_value_matrix('V_{!s}'.format(i),
                                                                    x=self._hdf_x,
                                                                    y=self._hdf_bias[i],
                                                                    unit='V',
                                                                    save_timestamp=False))
                if self._dVdI:
                    self._hdf_dVdI.append(self._data_file.add_value_matrix('dVdI_{!s}'.format(i),
                                                                           x=self._hdf_x,
                                                                           y=self._hdf_bias[i],
                                                                           unit='V/A',
                                                                           save_timestamp=False,
                                                                           folder='analysis',
                                                                           comment=self._get_numder_comment(self._hdf_V[i].name)+
                                                                                   '/'+self._get_numder_comment(self._hdf_I[i].name)))
                if self._fit_func:
                    self._hdf_fit.append(self._data_file.add_value_vector('{:s}_{:d}'.format(self._fit_name, i),
                                                                          x=self._hdf_x,
                                                                          unit=self._fit_unit,
                                                                          save_timestamp=False,
                                                                          folder='analysis',
                                                                          comment=self._get_fit_comment(i)))
                    self._data_fit.append(np.ones(len(self._x_vec)) * np.nan)
            # log-function
            self._add_log_value_vector()
            # add views
            self._add_views()
        elif self._scan_dim == 3:
            ''' 3D scan '''
            self._hdf_x = self._data_file.add_coordinate(self._x_coordname,
                                                         unit=self._x_unit)
            self._hdf_x.add(self._x_vec)
            self._hdf_y = self._data_file.add_coordinate(self._y_coordname,
                                                         unit=self._y_unit)
            self._hdf_y.add(self._y_vec)
            # add data variables
            self.sweeps.create_iterator()
            for i in range(self.sweeps.get_nos()):
                self._hdf_bias.append(self._data_file.add_coordinate('{:s}_b_{!s}'.format(self._IV_modes[self._bias], i),
                                                                     unit=self._IV_units[self._bias]))
                self._hdf_bias[i].add(self._get_bias_values(sweep=self.sweeps.get_sweep()))
                self._hdf_I.append(self._data_file.add_value_box('I_{!s}'.format(i),
                                                                 x=self._hdf_x,
                                                                 y=self._hdf_y,
                                                                 z=self._hdf_bias[i],
                                                                 unit='A',
                                                                 save_timestamp=False))
                self._hdf_V.append(self._data_file.add_value_box('V_{!s}'.format(i),
                                                                 x=self._hdf_x,
                                                                 y=self._hdf_y,
                                                                 z=self._hdf_bias[i],
                                                                 unit='V',
                                                                 save_timestamp=False))
                if self._dVdI:
                    self._hdf_dVdI.append(self._data_file.add_value_box('dVdI_{!s}'.format(i),
                                                                        x=self._hdf_x,
                                                                        y=self._hdf_y,
                                                                        z=self._hdf_bias[i],
                                                                        unit='V/A',
                                                                        save_timestamp=False,
                                                                        folder='analysis',
                                                                        comment=self._get_numder_comment(self._hdf_V[i].name)+
                                                                                '/'+self._get_numder_comment(self._hdf_I[i].name)))
                if self._fit_func:
                    self._hdf_fit.append(self._data_file.add_value_matrix('{:s}_{:d}'.format(self._fit_name, i),
                                                                          x=self._hdf_x,
                                                                          y=self._hdf_y,
                                                                          unit=self._fit_unit,
                                                                          save_timestamp=False,
                                                                          folder='analysis',
                                                                          comment=self._get_fit_comment(i)))
                    self._data_fit.append(np.ones((len(self._x_vec), len(self._y_vec))) * np.nan)
            # log-function
            self._add_log_value_vector()
            # add views
            self._add_views()
        ''' add comment '''
        if self._comment:
            self._data_file.add_comment(self._comment)
        return

    def _prepare_progress_bar(self):
        """
        Creates a progress bar using ipywidgets to show the measurement progress.
        Usually the number of performed sweeps, but in case of landscape scans entire number of bias points, is used as number of iterations.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        if self.progress_bar:
            # number of iterations as number of performed sweeps
            num_its = {0: len(self._x_vec),
                       1: self.sweeps.get_nos()*[1 if self._average is None else self._average][0],
                       2: len(self._x_vec)*self.sweeps.get_nos()*[1 if self._average is None else self._average][0],
                       3: len(self._x_vec)*len(self._y_vec)*self.sweeps.get_nos()*[1 if self._average is None else self._average][0]}
            if self._landscape and self._scan_dim in [2, 3]:
                # use entire number of bias points as number of iterations to estimate measurement time better
                bias_values = np.array([np.meshgrid(self._get_bias_values(sweep),
                                                    np.ones(len(self._x_vec)))[0]
                                        for sweep in self.sweeps.get_sweeps()])
                lsc_limits = np.array([self._lsc_vec,
                                        -self._lsc_vec if self._lsc_mirror
                                        else np.ones(shape=self._x_vec.shape)*np.array(self.sweeps.get_sweeps())[:,:2]])
                lsc_mask = np.logical_and(bias_values <= np.dot(np.ones((bias_values.shape[2], 1)), [np.max(lsc_limits, axis=0)]).T,
                                          bias_values >= np.dot(np.ones((bias_values.shape[2], 1)), [np.min(lsc_limits, axis=0)]).T)
                num_its[self._scan_dim] = np.sum(lsc_mask)*len(self._y_vec)*[1 if self._average is None else self._average][0]
                self._pb_addend = np.concatenate(list(zip(*np.sum(lsc_mask, axis=2))))  # value that counter has to be increased after each sweep
            self._pb = Progress_Bar(max_it=num_its[self._scan_dim],
                                    name='_'.join(list(filter(None, ('xy' if self._scan_dim is 0 else '{:d}D_IV_curve'.format(self._scan_dim), self._filename, self._expname)))))
        else:
            print('recording trace...')
    
    def _get_bias_values(self, sweep):
        """
        Gets a linearly distributed numpy-array of set bias values according to the sweep <sweep>.
        
        Parameters
        ----------
        sweep: array_likes of floats
            Sweep object containing start, stop, step and sleep values.
        
        Returns
        -------
        bias_values: numpy.array
            A linearly distributed numpy-array of given sweep parameters.
        """
        start = float(sweep[0])
        stop = float(sweep[1])
        step = float(sweep[2])
        ndigits = -int(np.floor(np.log10(np.abs(step)))) + 1
        nop = int(abs((stop - start) / step + np.sign(stop - start) * step / 100)) + 1
        return np.round(np.linspace(start, np.sign(stop - start) * (nop - 1) * step + start, nop), # stop is rounded down to multiples of step
                        decimals=ndigits)  # every element is rounded to overcome missing precision of numpy linspace
    
    def _get_numder_comment(self, name):
        """
        Save numerical derivative method (function, arguments, keyword arguments) to analysis comment in the .h5 file.
        
        Parameters
        ----------
        name: str
            Name of dataseries.
        
        Returns
        -------
        None
        """
        return '{!s}.{!s}('.format(self._numder_func.__module__, self._numder_func.__name__)+\
               ', '.join(np.concatenate([[name],
                                         np.array(self._numder_args, dtype=str) if self._numder_args else [],
                                         ['{:s}={!s}'.format(key, val) for key, val in self._numder_kwargs.items()] if self._numder_kwargs else []]))+\
               ')'

    def _get_fit_comment(self, i):
        """
        Save fit method (function, arguments, keyword arguments) to analysis comment in the .h5 file.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        return '{!s}.{!s}('.format(self._fit_func.__module__, self._fit_func.__name__)+\
               ', '.join(np.concatenate([['I={:s}'.format(self._hdf_I[i].name),
                                         'V={:s}'.format(self._hdf_V[i].name),
                                         'dVdI={:s}'.format(self._hdf_dVdI[i].name)],
                                        ['{:s}={!s}'.format(key, val) for key, val in self._fit_kwargs.items()] if self._fit_kwargs else []]))+\
               ')'

    def _add_log_value_vector(self):
        """
        Adds all value vectors for log-function parameter.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        if self.log_function != [None]:
            self._hdf_log = []
            self._data_log = []
            for i, _ in enumerate(self.log_function):
                if self._scan_dim == 1:
                    self._hdf_log.append(self._data_file.add_coordinate(self.log_name[i],
                                                                        unit=self.log_unit[i]))
                    self._data_log.append(np.nan)
                elif self._scan_dim == 2:
                    self._hdf_log.append(self._data_file.add_value_vector(self.log_name[i],
                                                                          x=self._hdf_x,
                                                                          unit=self.log_unit[i],
                                                                          dtype=self.log_dtype[i]))
                    self._data_log.append(np.ones(len(self._x_vec))*np.nan)
                elif self._scan_dim == 3:
                    self._hdf_log.append(self._data_file.add_value_matrix(self.log_name[i],
                                                                          x=self._hdf_x,
                                                                          y=self._hdf_y,
                                                                          unit=self.log_unit[i],
                                                                          dtype=self.log_dtype[i]))
                    self._data_log.append(np.ones((len(self._x_vec), len(self._y_vec)))*np.nan)
        return
    
    def _add_views(self):
        """
        Adds views to the .h5-file. The view "IV" plots I(V) and contains the whole set of sweeps that are set.
        If <dVdI> is true, the view "dVdI" plots the differential gradient dV/dI(V) and contains the whole set of sweeps that are set.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        self._hdf_view_IV = self._data_file.add_view('IV',
                                                     x=self._hdf_V[0],
                                                     y=self._hdf_I[0],
                                                     view_params={"labels": ('V', 'I'),
                                                                  'plot_style': 1,
                                                                  'markersize': 5})
        for i in range(1, self.sweeps.get_nos()):
            self._hdf_view_IV.add(x=self._hdf_V[i], y=self._hdf_I[i])
        if self._dVdI:
            self._hdf_view_dVdI = self._data_file.add_view('dVdI',
                                                           x=self._hdf_I[0],
                                                           y=self._hdf_dVdI[0],
                                                           view_params={"labels": ('I', 'dVdI'),
                                                                        'plot_style': 1,
                                                                        'markersize': 5})
            for i in range(1, self.sweeps.get_nos()):
                self._hdf_view_dVdI.add(x=[self._hdf_I, self._hdf_V][self._bias][i],
                                        y=self._hdf_dVdI[i])
        return
    
    def _get_sweepdata(self):
        """
        Iterates sweeps of sweep class and takes data for each sweep.
        If average is set, traces are taken <average>-fold to average and saved after each iteration
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        if self._average is None:
            self.sweeps.create_iterator()
            for j in range(self.sweeps.get_nos()):
                # take data
                I_values, V_values = self.take_IV(sweep=self.sweeps.get_sweep())
                data = {self._hdf_I[j]:(I_values,),
                        self._hdf_V[j]:(V_values,)} # tuple in oder to use *args later
                if self._dVdI:
                    data[self._hdf_dVdI[j]] = (self._numerical_derivative(I_values, V_values),)
                if self._fit_func:
                    #data[self._hdf_fit[j]] = self._fit_func(data[self._hdf_I[j]],
                    #                                        data[self._hdf_V[j]],
                    #                                        data[self._hdf_dVdI[j]],
                    #                                        **self._fit_kwargs)
                    if self._scan_dim == 1:
                        self._data_fit[j] = float(self._fit_func(data[self._hdf_I[j]][0],
                                                                           data[self._hdf_V[j]][0],
                                                                           data[self._hdf_dVdI[j]][0],
                                                                           **self._fit_kwargs))
                        data[self._hdf_fit[j]] = (self._data_fit[j],)
                    elif self._scan_dim == 2:
                        self._data_fit[j][self.ix] = float(self._fit_func(data[self._hdf_I[j]][0],
                                                                          data[self._hdf_V[j]][0],
                                                                          data[self._hdf_dVdI[j]][0],
                                                                          **self._fit_kwargs))
                        data[self._hdf_fit[j]] = (self._data_fit[j], True)
                    elif self._scan_dim == 3:
                        self._data_fit[j][self.ix, self.iy] = float(self._fit_func(data[self._hdf_I[j]][0],
                                                                                   data[self._hdf_V[j]][0],
                                                                                   data[self._hdf_dVdI[j]][0],
                                                                                   **self._fit_kwargs))
                        data[self._hdf_fit[j]] = (self._data_fit[j][self.ix], self._rst_fit_hdf_appnd)
                        self._rst_fit_hdf_appnd = not bool(self.iy + 1 == len(self._y_vec))
                # save data
                for key, val in data.items():
                    key.append(*val)
                # iterate progress bar
                if self.progress_bar:
                    self._pb.iterate(addend=self._pb_addend[self.ix] if self._landscape else 1)
                qkit.flow.sleep()
        else:
            I_values, V_values = [], []
            for i in range(self._average):
                I_values.append([])
                V_values.append([])
                self.sweeps.create_iterator()
                for j in range(self.sweeps.get_nos()):
                    # take data
                    for val, lst in zip(self.take_IV(sweep=self.sweeps.get_sweep()), [I_values, V_values]):
                        lst[i].append(list(val))  # append as list in order to later use zip
                    I_values_avg, V_values_avg = np.mean(list(zip(*I_values))[j], axis=0), np.mean(list(zip(*V_values))[j], axis=0)  # use zip since np.mean cannot handle different shapes
                    data = {self._hdf_I[j]:I_values_avg,
                            self._hdf_V[j]:V_values_avg}
                    if self._dVdI:
                        data[self._hdf_dVdI[j]] = self._numerical_derivative(I_values_avg, V_values_avg)
                    if self._fit_func:
                        data[self._hdf_fit[j]] = self._fit_func(data[self._hdf_I[j]],
                                                                data[self._hdf_V[j]],
                                                                data[self._hdf_dVdI[j]],
                                                                **self._fit_kwargs)
                    # save data
                    for key, val in data.items():
                        key.append(val, reset=bool(i))  # append data series or overwrite last iteration by new averaged data
                        key.ds.attrs['average'] = '({:d}/{:d})'.format(i+1, self._average)  # add (iteration/average) as attribute
                        self._data_file.flush()
                    # iterate progress bar
                    if self.progress_bar:
                        self._pb.iterate(addend=self._pb_addend[self.ix] if self._landscape else 1)
            # set average attribute to number of averages
            for j in range(self.sweeps.get_nos()):
                for lst in [val for k, val in enumerate([self._hdf_I, self._hdf_V, self._hdf_dVdI]) if k < 2+int(self._dVdI)]:
                    lst[j].ds.attrs['average'] = self._average
            self._data_file.flush()
            qkit.flow.sleep()
        return
    
    def take_IV(self, sweep):
        """
        Takes IV and considers if landscape is set.
        
        Parameters
        ----------
        sweep: array_likes of floats
            Sweep object containing start, stop, step and sleep values.
        
        Returns
        -------
        I_values: numpy.array(float)
            Measured current values.
        V_values: numpy.array(float)
            Measured voltage values.
        """
        # take data
        if self._landscape:
            # modify sweep by envelop of landscape function
            bias_lim = self._lsc_vec[self.ix]
            if self._lsc_mirror:
                sweep_lsc = np.nanmin([np.abs(sweep), [bias_lim, bias_lim, np.nan, np.nan]], axis=0)*np.sign(sweep)
            else:
                sweep_lsc = np.nanmin([sweep, [bias_lim, bias_lim, np.nan, np.nan]], axis=0)
            # find landscape bounds in full bias values
            bias_data = self._get_bias_values(sweep)
            mask = np.logical_and(bias_data >= np.min(sweep_lsc[:2]), bias_data <= np.max(sweep_lsc[:2]))
            sweep_lsc[:2] = bias_data[np.where(mask)[0]][0], bias_data[np.where(mask)[0]][-1]
            # take data
            data = self._IVD.take_IV(sweep=sweep_lsc)
            # fill skipped bias values with np.nan to keep shape constant
            I_values, V_values = np.array([np.nan] * len(mask)), np.array([np.nan] * len(mask))
            np.place(arr=I_values, mask=mask, vals=data[0])
            np.place(arr=V_values, mask=mask, vals=data[1])
        else:
            I_values, V_values = self._IVD.take_IV(sweep=sweep)
        time.sleep(sweep[3])
        return I_values, V_values
    
    def set_plot_comment(self, comment):
        """
        Set small comment to add at the end of plot pics for more information i.e. good for wiki entries.
        
        Parameters
        ----------
        comment: str
            A comment that is added in the plot images.
        
        Returns
        -------
        None
        """
        self._plot_comment = comment
        return
    
    class sweep(object):
        """
        This is a subclass of <transport> that provides the customized usage of many sweeps in one measurement.
        """
        def __init__(self):
            """
            Initializes the sweep parameters as empty.
        
            Parameters
            ----------
            None
        
            Returns
            -------
            None
            """
            self._starts = []
            self._stops = []
            self._steps = []
            self._sleeps = []
            self.create_iterator()
            return
        
        def create_iterator(self):
            """
            Creates iterator of start, stop, step and sleep arrays.
        
            Parameters
            ----------
            None
        
            Returns
            -------
            None
            """
            self._start_iter = iter(self._starts)
            self._stop_iter = iter(self._stops)
            self._step_iter = iter(self._steps)
            self._sleep_iter = iter(self._sleeps)
            return
        
        def add_sweep(self, start, stop, step, sleep=0):
            """
            Adds a sweep object with given parameters.
        
            Parameters
            ----------
            start: float
                Start value of the sweep.
            stop: float
                Stop value of the sweep.
            step: float
                Step value of the sweep.
            sleep: float
                Sleep time after whole sweep. Default is 0.
        
            Returns
            -------
            None
            """
            self._starts.append(start)
            self._stops.append(stop)
            self._steps.append(step)
            self._sleeps.append(sleep)
            return
        
        def reset_sweeps(self):
            """
            Resets sweeps.
        
            Parameters
            ----------
            None
        
            Returns
            -------
            None
            """
            self._starts = []
            self._stops = []
            self._steps = []
            self._sleeps = []
            return
        
        def get_sweep(self):
            """
            Gets sweep parameter.
        
            Parameters
            ----------
            None
        
            Returns
            -------
            sweep: tuple of floats
                Sweep object containing start, stop, step and sleep values.
            """
            return (next(self._start_iter),
                    next(self._stop_iter),
                    next(self._step_iter),
                    next(self._sleep_iter))
        
        def get_sweeps(self):
            """
            Gets parameters of all sweep.
        
            Parameters
            ----------
            None
        
            Returns
            -------
            sweep: list of floats
                Sweep object containing start, stop, step and sleep values.
            """
            return list(zip(*[self._starts, self._stops, self._steps, self._sleeps]))
        
        def get_nos(self):
            """
            Gets number of sweeps.
        
            Parameters
            ----------
            None
        
            Returns
            -------
            nos: int
                Number of sweeps stored in this sweep-class.
            """
            return len(self._starts)



# This instrument transmits and receives raw ethernet packets to and from 
# a network interface. It is used to communicate with GHZdac devices.
# Markus Jerger <jerger@kit.edu>, 2011
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# TO DO
# - cleanup/timeout of registered listeners
#
#

import types
import logging
try:
    from qkit.core.instrument_base import Instrument
except:
    # debug mode
    logging.getLogger().setLevel(logging.INFO)
    logging.warning(__name__ + ': unable to load instrument module. entering standalone mode.')
    class Instrument:
        FLAG_SET = 0x01
        FLAG_GET = 0x02
        FLAG_GETSET = FLAG_SET | FLAG_GET
        def __init__(self, name, tags):
            pass
        def add_parameter(self, name, type, flags, channels = (), minval = None, maxval = None, units = None, format = None, tags = None):
            if(flags & self.FLAG_GET): setattr(self, 'get_%s'%name, getattr(self, 'do_get_%s'%name))
            if(flags & self.FLAG_SET): setattr(self, 'set_%s'%name, getattr(self, 'do_set_%s'%name))
        def add_function(self, func):
            pass
# comms and byte packing
import threading
import socket
if(not hasattr(socket, 'PF_PACKET')): socket.PF_PACKET = 17
import dpkt
import struct
import winpcapy as capy
import ctypes

# ip filter
import re

def iflist():
    '''
    return a list of devices supported by winpcap
    '''
    alldevs = ctypes.POINTER(capy.pcap_if_t)()
    d = ctypes.POINTER(capy.pcap_if_t)
    errbuf = ctypes.create_string_buffer(capy.PCAP_ERRBUF_SIZE)

    ## Retrieve the device list
    if (capy.pcap_findalldevs(ctypes.byref(alldevs), errbuf) == -1):
        raise Exception("Error in retrieveing interface list: %s\n" % errbuf.value)
        return

    devices = []
    d=alldevs.contents
    while d:
        devices.append({'name': d.name, 'description': d.description})
        if d.next:
             d=d.next.contents
        else:
             d=False
    ## Free the device list
    capy.pcap_freealldevs(alldevs)
    
    return devices

class EthernetServerWin(Instrument):
    '''
    This instrument passes packets between UDP and Ethernet, to talk to GHZdac devices.
    It uses winpcap to capture and transmit packets in windows.

    Usage:
    Initialize with
    <name> = qt.instruments.create('name', 'EthernetServer', interface='<interface name>', MAC='<MAC byte string>', ethertype=<ethertype>, ip_host=<ip_host>, ip_port=<ip_port>)
    <interface name> = name of the raw packet network interface
        interface='\\Device\\NPF_{FB9A232D-284F-4039-A1BA-2B59FBE33AF2}' on pi-us20
    <MAC> = MAC address of the raw interface, MAC='00:26:18:D9:93:F4' on pi-us20
    <ethertype> = 0x0003 (ETH_P_ALL, receive all packets) or any other EtherType value
    <ip_host>:<ip_port> = 127.0.0.1:413 to bind the UDP interface to
    '''

    # broadcast address
    MAC_BROADCAST = struct.pack('>6B', *[0xff, 0xff, 0xff, 0xff, 0xff, 0xff])
    # capture all packets ethertype
    ETH_P_ALL = 0x0003
    # maximum packet size
    RECV_SIZE = 2048

    def __init__(self, name, interface, MAC, ethertype = 0x0003, ip_host = '127.0.0.1', ip_port = 413, ip_filter = ''):
        '''
        Initializes the UDP-to-Ethernet bridge

        Input:
            name (string)    : name of the instrument
            interface (string) : raw ethernet network interface
            MAC (string) : MAC address of the interface in hex, bytes separated by :
            ethertype : ethernet frame type to receive
            ip_host:ip_port : UDP interface address
            ip_filter : regular expression used to match incoming packets against

        Output:
            None
        '''
        logging.debug(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name, tags=['physical'])

        self._interface = interface
        self._MAC = struct.pack('>6B', *[int(x, 16) for x in MAC.split(':')])
        self._ethertype = int(ethertype)
        self._ip_host = ip_host
        self._ip_port = int(ip_port)
        self.do_set_ip_filter(ip_filter)
        self._clients = {} # associate MAC addresses with udp clients
        self._devices = set() # mac addresses seen on the raw interface
        self._updateLock = threading.Lock()

        # add parameters
        self.add_parameter('interface', type=str, flags=Instrument.FLAG_GET)
        self.add_parameter('MAC', type=str, flags=Instrument.FLAG_GET)
        self.add_parameter('ethertype', type=int, flags=Instrument.FLAG_GET)
        self.add_parameter('ip_host', type=str, flags=Instrument.FLAG_GET)
        self.add_parameter('ip_port', type=int, flags=Instrument.FLAG_GET)
        self.add_parameter('ip_filter', type=str, flags=Instrument.FLAG_GETSET)
        self.add_parameter('devices', type=types.ListType, flags=Instrument.FLAG_GET)
        self.add_parameter('listeners', type=types.ListType, flags=Instrument.FLAG_GET)
        self.add_parameter('raw_rcvbuf', type=int, flags=Instrument.FLAG_GETSET)
        self.add_parameter('udp_rcvbuf', type=int, flags=Instrument.FLAG_GETSET)
        self.add_function('get_all')

        # open sockets
        self._init_udp()
        self._init_raw()

        # update ui
        self.get_all()

        # start daemons
        self._udp_receiver = self.udp_receiver(self)
        self._udp_receiver.daemon = True
        self._udp_receiver.start()
        self._raw_receiver = self.raw_receiver(self)
        self._raw_receiver.daemon = True
        self._raw_receiver.start()

    def __del__(self):
        '''
        clean up
        '''
        self._udp_socket.close()
        self._raw_socket.close()

    class raw_socket:
        '''
            a socket-like raw send/receive interface built around winpcapy
        '''
        _so_rcvbuf = 8192
        _so_snaplen = 65536 # max length supported by ethernet length descriptor
        _so_timeout = 1000 # 1s
        _so_promicious = 0
        
        def __init__(self, family, type, proto):
            '''
            initialize socket
            '''
            if((family != socket.PF_PACKET) or (type != socket.SOCK_RAW)):
                raise Exception('raw_socket supports only family=PF_PACKET, type=SOCK_RAW.')
            self._ethertype = proto
            # create libpcap data structures
            self._errbuf = ctypes.create_string_buffer(capy.PCAP_ERRBUF_SIZE)
            self._fp = capy.pcap_t
        
        def close():
            '''
            free resources
            '''
            capy.pcap_close(self._fp)
            
        def bind(self, tuple):
            '''
            bind socket to a given interface/packet type pair
            '''
            #fp = pcap_open_live(sys.argv[1],65536,1,1000,errbuf)
            self._interface = tuple[0]
            self._ethertype = tuple[1]
            self._fp = capy.pcap_open_live(self._interface, self._so_snaplen, self._so_promicious, self._so_timeout, self._errbuf)
            if not bool(self._fp):
                raise Exception('unable to open interface %s for packet capture.'%self._interface)
            
        def getsockname(self):
            '''
            '''
            return [False, False, False, False, '']
        
        def setsockopt(self, level, option, value):
            '''
            set socket options. supports only level=socket.SOL_SOCKET option=socket.SO_RCVBUF.
            '''
            pass

        def getsockopt(self, level, option):
            '''
            get socket options. supports only level=socket.SOL_SOCKET option=socket.SO_RCVBUF.
            '''
            pass

        def send(self, data):
            '''
            send string data out as a raw packet
            '''
            #packet = (ctypes.c_ubyte * len(data))
            #for i in range(len(data)): packet[i] = data[i] # todo: nicer solution
            packet = ctypes.cast(ctypes.c_char_p(data), ctypes.POINTER(ctypes.c_ubyte))
            capy.pcap_sendpacket(self._fp, packet, len(data))
            
        def recvfrom(self, bufsize):
            # retrieve a packet from pcap
            header = ctypes.POINTER(capy.pcap_pkthdr)()
            pkt_data = ctypes.POINTER(ctypes.c_ubyte)()
            status = capy.pcap_next_ex(self._fp, ctypes.byref(header), ctypes.byref(pkt_data))
            if(status == 0):
                # timeout
                return False
            elif(status > 0):
                # interpret captured packet
                pkt_len = header.contents.caplen
                pkt_buf = ctypes.string_at(pkt_data, pkt_len)
                pkt = dpkt.ethernet.Ethernet(pkt_buf)
                return (pkt_buf, pkt.src)
            else:
                # some error occured
                raise('error capturing packets: %s'%capy.pcap_geterr(self._fp))
            pass
            

    class udp_receiver(threading.Thread):
        '''
        receive client requests
        '''
        def __init__(self, node):
            threading.Thread.__init__(self, name = 'node_udp')
            self.node = node

        def run(self):
            logging.debug(self.name + ' : udp listener started.')
            while(True):
                # receive packet (ethernet packet in the payload of an UDP packet)
                (raw_packet, sender) = self.node._udp_socket.recvfrom(self.node.RECV_SIZE)
                packet = dpkt.ethernet.Ethernet(raw_packet)
                logging.debug(self.name + ' : udp packet received.')
                # source filtering
                if(not self.node._ip_filter(sender)):
                    logging.debug(self.name + ' : udp packet from %s discarded.'%sender[0])
                    continue
                # add this connection to the listeners of a specific device MAC
                if(packet.dst not in self.node._clients):
                    self.node._updateLock.acquire(True)
                    if(packet.dst not in self.node._clients):
                        self.node._clients[packet.dst] = set()
                    self.node._updateLock.release()
                if(sender not in self.node._clients[packet.dst]):
                    self.node._clients[packet.dst].add(sender)
                # dispatch to raw socket
                packet.src = self.node._MAC
                self.node._raw_socket.send(packet.pack())

    class raw_receiver(threading.Thread):
        '''
        receive raw network packets
        '''
        def __init__(self, node):
            threading.Thread.__init__(self, name = 'node_raw')
            self.node = node

        def run(self):
            logging.debug(self.name + ' : raw listener started.')
            while(True):
                # receive packet (ethernet packet)
                (raw_packet, sender) = self.node._raw_socket.recvfrom(self.node.RECV_SIZE)
                packet = dpkt.ethernet.Ethernet(raw_packet)
                logging.debug(self.name + ' : raw packet received.')
                self.node._devices.add(packet.src)
                # dispatch to all listeners
                for client_list in [self.node._clients.get(packet.src, None), self.node._clients.get(self.node.MAC_BROADCAST, None)]:
                    if(client_list == None): continue
                    for client in client_list:
                        self.node._udp_socket.sendto(raw_packet, client)

    def _init_udp(self):
        # open udp socket
        logging.debug(__name__ + ' : Listening for packets on %s:%d.'%(self._ip_host, self._ip_port))
        self._udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._udp_socket.bind((self._ip_host, self._ip_port))

    def _init_raw(self):
        # open raw socket
        if(self._ethertype == self.ETH_P_ALL):
            logging.info(__name__ + ' : EtherType set to ETH_P_ALL, this may generate high CPU load')
        self._raw_socket = self.raw_socket(socket.PF_PACKET, socket.SOCK_RAW, self._ethertype)
        self._raw_socket.bind((self._interface, self._ethertype))
        #self._MAC = self._raw_socket.getsockname()[4] # currently unsupported

    def get_all(self):
        ''' call all getters to update the ui '''
        self.get_interface()
        self.get_MAC()
        self.get_ethertype()
        self.get_ip_host()
        self.get_ip_port()
        self.get_ip_filter()
        self.get_udp_rcvbuf()
        self.get_raw_rcvbuf()

        self.get_listeners()
        self.get_devices()

    def do_get_interface(self):
        ''' return name of the network interface used '''
        return self._interface

    def do_get_MAC(self):
        ''' return own MAC address '''
        return '%02x:%02x:%02x:%02x:%02x:%02x'%(struct.unpack('>6B', self._MAC))

    def do_get_ethertype(self):
        ''' return packet type received '''
        return self._ethertype

    def do_get_ip_host(self):
        ''' return hostname of the udp socket '''
        return self._ip_host

    def do_get_ip_port(self):
        ''' return port number of the udp socket '''
        return self._ip_port

    def do_set_ip_filter(self, expr):
        ''' udp source filter (regular) expression '''
        if(expr):
            self._ip_filter_expr = expr
            self._ip_filter = re.compile(expr).match
        else:
            self._ip_filter_expr = ''
            self._ip_filter = lambda s: True

    def do_get_ip_filter(self):
        ''' return udp source filter expression '''
        return self._ip_filter_expr

    def do_get_listeners(self):
        ''' return list of udp clients '''
        # gather list of clients in <ip>:<port> format
        listeners = []
        for client_list in self._clients.values():
            listeners.append(['%s:%d'%(x[0], x[1]) for x in list(client_list)])
        # remove duplicates
        return list(set(listeners))

    def do_get_devices(self):
        ''' return list of seen devices '''
        #return self._clients.keys()
        return list(self._devices)

    def do_set_raw_rcvbuf(self, bufsize):
        ''' set the receive buffer size of the raw socket '''
        self._raw_socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, int(bufsize))

    def do_get_raw_rcvbuf(self):
        ''' get the receive buffer size of the raw socket '''
        return self._raw_socket.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF)

    def do_set_udp_rcvbuf(self, bufsize):
        ''' set the receive buffer size of the raw socket '''
        self._udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, int(bufsize))

    def do_get_udp_rcvbuf(self):
        ''' get the receive buffer size of the raw socket '''
        return self._udp_socket.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF)




# MP, AS @ KIT 05/2017
# Class to collect all information about a single measurement

import copy
import json
import logging
import os

import qkit
from qkit.measure.json_handler import QkitJSONEncoder, QkitJSONDecoder
from qkit.measure.samples_class import Sample


class Measurement(object):
    '''
    Measurement class to define general information about the specific measurement.
    '''
    def __init__(self):
        try:
            self.git_commit_id = qkit.git['commit_id']
        except AttributeError:
            pass
        self.uuid = None
        self.hdf_relpath = ''
        self.sample = None
        self.web_visible = True
        self.rating = 3
        self.analyzed = False
        self.measurement_type = ''
        self.measurement_func = ''
        self.x_axis = ''
        self.y_axis = ''
        self.z_axis = ''
        self.instruments = None
        self.user = qkit.cfg.get('user','John_Doe').strip().replace(" ","_")
        self.run_id = qkit.cfg.get('run_id','NO_RUN').strip().replace(" ","_").upper()

    def get_JSON(self):
        """
        Returns JSON string of all attributes. The function is mostly used for saving in .h5.
        """
        return json.dumps(obj=self._get_copydict(), cls=QkitJSONEncoder, indent = 4, sort_keys=True)

    def save(self,filepath=None):
        '''
        Save sample object in the same directory as the measurement .h5 file or in the given filepath.
        '''
        if filepath:
            filepath=filepath
        else:
            filepath = os.path.join(qkit.cfg.get('datadir'),self.hdf_relpath.replace('.h5', '.measurement'))

        with open(filepath,'w+') as filehandler:
            json.dump(obj=self._get_copydict(), fp=filehandler, cls=QkitJSONEncoder, indent = 4, sort_keys=True)

    def load(self, filename):
        '''
        Load sample keys and entries to current sample instance.
        '''
        if not os.path.isabs(filename):
            filename = os.path.join(qkit.cfg.get('datadir'),filename)

        with open(filename) as filehandle:
            self.__dict__ = json.load(filehandle, cls = QkitJSONDecoder)

        s = Sample()
        s.__dict__ = self.__dict__['sample']
        self.__dict__['sample'] = s
    
    def _get_copydict(self):
        '''
        Returns a copy of self.__dict__ with all object attributes returned in a JSONable format.
        '''
        copydict = copy.copy(self.__dict__)

        """
        Both 'sample' and 'instruments' entries have to be converted into a dict to save information about datatype etc.
        """
        if copydict['sample']:
            copydict['sample'] = copydict['sample'].__dict__ # change entry to dict to make it readable for JSONEncoder
        if copydict['instruments']:
            if not isinstance(copydict['instruments'], dict): # conversion needed during measurement
                copydict['instruments'] = self._JSON_instruments_dict()
        return copydict
            
    def _JSON_instruments_dict(self):
        '''
        Iterates through all entries in the self.instruments dict and creates a dict with the individial instrument
        name together with all attributs, attribute values and information about setter function. This is needed
        for an automized reading and re-setting of a measurement.
        '''
        return_dict = {}
        for ins_name in self.instruments:
            ins = qkit.instruments.get(ins_name)
            param_dict = {}
            for param_name in ins.get_parameter_names():
                has_setter = 'set_func' in ins.get_parameter_options(param_name)
                param_dict.update({param_name:{'content':ins.get(param_name, query=False), 'has_setter':has_setter}})
            return_dict.update({ins_name:param_dict})
        return return_dict

    def update_instrument(self, ins_name):
        '''
        Sets all parameters of the instrument ins_name to the values used in the measurements.
        '''
        try:
            ins = qkit.instruments.get(ins_name)
            params_dict = self.instruments['instruments'][ins_name]
        except AttributeError or NameError:
                logging.error('Relevant instruments and attributes not properly specified. '+str(ins_name))
        else:
            '''
            Use only parameters with setter and update the instrument
            '''
            for param in params_dict.keys():
                if params_dict[param]['has_setter']: ins.set(params_dict[param], params_dict[param]['content'])
    
    def update_all_instruments(self):
        '''
        Updates all instruments to their parameter settings during the measurement.
        CAUTION: This can be critical for current sources, since the current value is set and not ramped!
        '''
        for ins in self.instuments:
            self.update_instrument(ins)

    def write_to_hdf(self, hdf_file):
        self.uuid = hdf_file._uuid
        self.hdf_relpath = hdf_file._relpath
        mo = hdf_file.add_textlist('measurement')
        mo.append(self.get_JSON())



# qtflow.py, handle 'flow control' in the QT lab environment
# Pieter de Groot, <pieterdegroot@gmail.com>, 2008
# Reinier Heeres, <reinier@heeres.eu>, 2008
# HR@KIT/2017 (python3 conversion & cleanup for QKIT)
# YS@KIT/2018: Further emancipated from qtlab
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import logging
import time
import qkit
from qkit.core.lib.misc import get_traceback

AutoFormattedTB = get_traceback()

class FlowControl(object):
    '''
    Class for flow control of the QT measurement environment.
    '''

    STATUS_STOPPED = 0
    STATUS_RUNNING = 1

    def __init__(self):
        
        self._status = 'starting'
        self._measurements_running = 0
        self._abort = False
        self._pause = False
        self._exit_handlers = []
        self._callbacks = {}

    #########
    ### signals
    #########

    def measurement_start(self):
        '''
        Indicate the start of a measurement.

        FIXME: The following is disabled due to lack of exception catching.
        This will increment the internal measurement counter, and if a
        measurement was not running, it will emit the 'measurement-start'
        signal.
        '''

        self._measurements_running += 1
        if self._measurements_running == 1:
            self._set_status('running')

            # Handle callbacks
            self.run_mainloop(1, wait=False)

    def measurement_end(self, abort=False):
        '''
        Indicate the end of a measurement.

        FIXME: The following is disabled due to lack of exception catching.
        This will decrement the internal measurement counter if abort=False,
        and set it to 0 in abort=True. If the counter reached zero (e.g. the
        last measurement was stopped, it will emit the 'measurement-end'
        signal.
        '''

        if abort:
            self._measurements_running = 0
        elif self._measurements_running > 0:
            self._measurements_running -= 1

        if self._measurements_running == 0:
            self._set_status('stopped')

            # Handle callbacks
            self.run_mainloop(1, wait=False)

    def run_mainloop(self, delay, wait=True):
        '''
        Run mainloop for a maximum of <delay> seconds.
        If wait is True (default), sleep until <delay> seconds have passed.
        '''
        if wait and delay > 0:
            time.sleep(delay)

    def measurement_idle(self, delay=0.0):
        '''
        Indicate that the measurement is idle and handle events.

        This function will check whether an abort has been requested and
        handle queued events for a time up to 'delay' (in seconds).
        
        Be aware that both time.time and time.sleep have a granularity of around 15ms on windows.
        However, this should not be limiting.

        It starts by emitting the 'measurement-idle' signal to allow callbacks
        to be executed by the time this function handles the event queue.
        After that it handles events and sleeps for periods of 10msec. Every
        <emit_interval> seconds it will emit another measurement-idle signal.
        '''
        start = time.time()

        while self._pause:
            self.check_abort()
            self.run_mainloop(0.01,wait=True)

        while True:
            self.check_abort()

            dt = time.time() - start
            self.run_mainloop(delay-dt, wait=False)

            dt = time.time() - start
            if dt + 0.02 < delay:
                time.sleep(0.05)
            else:
                time.sleep(max(0, delay - dt))
                return

    def _run_script(self, scriptfile):
        return execfile(scriptfile)

    def register_exit_script(self, scriptfile):
        self.register_exit_handler(lambda: self._run_script(scriptfile))

    def register_exit_handler(self, func):
        if func not in self._exit_handlers:
            self._exit_handlers.append(func)

    def exit_request(self):
        '''Run all registered exit handlers.'''
        for func in self._exit_handlers:
            try:
                func()
            except Exception as e:
                print('Error in func %s: %s' % (func.__name__, str(e)))


    ############
    ### status
    ############

    def get_status(self):
        '''Get status, one of "running", "stopped", "starting" '''
        return self._status

    def _set_status(self, val):
        self._status = val

    def finished_starting(self):
        self._status = "stopped"

    def is_measuring(self):
        return self.get_status() == 'running'

    def check_abort(self):
        '''Check whether an abort has been requested.'''

        if self._abort:
            self._abort = False
            self.measurement_end(abort=True)
            raise ValueError('Human abort')

    def set_abort(self):
        '''Request an abort.'''
        self._abort = True

    def is_paused(self):
        return self._pause

    def set_pause(self, pause):
        '''Set / unset pause state.'''
        self._pause = pause

def exception_handler(self, etype, value, tb, tb_offset=None):
    # when the 'tb_offset' keyword argument is omitted above, ipython
    # raises an error.
    TB = AutoFormattedTB(mode='Context', color_scheme='Linux', tb_offset=1)

    fc = qkit.flow

    # for the unlikely event that an error occured after pushing the
    # stop button
    fc._abort = False

    if fc.is_measuring():
        # put qtlab back in 'stopped' state
        fc.measurement_end(abort=True)

    TB(etype, value, tb)

def qtlab_exit():
    print("Closing QKIT...")
